from datetime import datetime
import pytest
from nylas.client.restful_models import Account, APIAccount, SingletonAccount


def test_create_account(api_client, monkeypatch):
    monkeypatch.setattr(api_client, "is_opensource_api", lambda: False)
    account = api_client.accounts.create()
    assert isinstance(account, Account)


def test_create_apiaccount(api_client, monkeypatch):
    monkeypatch.setattr(api_client, "is_opensource_api", lambda: True)
    account = api_client.accounts.create()
    assert isinstance(account, APIAccount)


def test_account_json(api_client, monkeypatch):
    monkeypatch.setattr(api_client, "is_opensource_api", lambda: False)
    account = api_client.accounts.create()
    result = account.as_json()
    assert isinstance(result, dict)


@pytest.mark.usefixtures("mock_ip_addresses")
def test_ip_addresses(api_client_with_client_id):
    result = api_client_with_client_id.ip_addresses()
    assert isinstance(result, dict)
    assert "updated_at" in result
    assert "ip_addresses" in result


@pytest.mark.usefixtures("mock_token_info", "mock_account")
def test_token_info(api_client_with_client_id):
    result = api_client_with_client_id.token_info()
    assert isinstance(result, dict)
    assert "updated_at" in result
    assert "scopes" in result


@pytest.mark.usefixtures("mock_account")
def test_account_datetime(api_client):
    account = api_client.account
    assert account.linked_at == datetime(2017, 7, 24, 18, 18, 19)


@pytest.mark.usefixtures("mock_accounts", "mock_account_management")
def test_account_upgrade(api_client, client_id):
    api_client.client_id = client_id
    account = api_client.accounts.first()
    assert account.billing_state == "paid"
    account = account.downgrade()
    assert account.billing_state == "cancelled"
    account = account.upgrade()
    assert account.billing_state == "paid"


@pytest.mark.usefixtures("mock_revoke_all_tokens", "mock_account")
def test_revoke_all_tokens(api_client_with_client_id):
    assert api_client_with_client_id.access_token is not None
    api_client_with_client_id.revoke_all_tokens()
    assert api_client_with_client_id.access_token is None


@pytest.mark.usefixtures("mock_revoke_all_tokens", "mock_account")
def test_revoke_all_tokens_with_keep_access_token(
    api_client_with_client_id, access_token
):
    assert api_client_with_client_id.access_token == access_token
    api_client_with_client_id.revoke_all_tokens(keep_access_token=access_token)
    assert api_client_with_client_id.access_token == access_token


@pytest.mark.usefixtures("mock_accounts", "mock_account")
def test_account_access(api_client):
    account1 = api_client.account
    assert isinstance(account1, SingletonAccount)
    account2 = api_client.accounts[0]
    assert isinstance(account2, APIAccount)
    account3 = api_client.accounts.first()
    assert isinstance(account3, APIAccount)
    assert account1.as_json() == account2.as_json() == account3.as_json()


@pytest.mark.usefixtures("mock_accounts")
def test_account_metadata(api_client_with_client_id, monkeypatch):
    monkeypatch.setattr(api_client_with_client_id, "is_opensource_api", lambda: False)
    account1 = api_client_with_client_id.accounts[0]
    account1["metadata"] = {"test": "value"}
    account1.save()
    assert account1["metadata"] == {"test": "value"}


@pytest.mark.usefixtures("mock_accounts")
def test_application_account_delete(api_client_with_client_id, monkeypatch):
    monkeypatch.setattr(api_client_with_client_id, "is_opensource_api", lambda: False)
    account1 = api_client_with_client_id.accounts[0]
    api_client_with_client_id.accounts.delete(account1.id)


@pytest.mark.usefixtures("mock_application_details")
def test_application_details(api_client_with_client_id, monkeypatch):
    monkeypatch.setattr(api_client_with_client_id, "is_opensource_api", lambda: False)
    app_data = api_client_with_client_id.application_details()
    assert app_data["application_name"] == "My New App Name"
    assert app_data["icon_url"] == "http://localhost:5555/icon.png"
    assert app_data["redirect_uris"] == [
        "http://localhost:5555/login_callback",
        "localhost",
        "https://customerA.myapplication.com/login_callback",
    ]


@pytest.mark.usefixtures("mock_application_details")
def test_update_application_details(api_client_with_client_id, monkeypatch):
    monkeypatch.setattr(api_client_with_client_id, "is_opensource_api", lambda: False)
    updated_data = api_client_with_client_id.update_application_details(
        application_name="New Name",
        icon_url="https://myurl.com/icon.png",
        redirect_uris=["https://redirect.com"],
    )
    assert updated_data["application_name"] == "New Name"
    assert updated_data["icon_url"] == "https://myurl.com/icon.png"
    assert updated_data["redirect_uris"] == [
        "https://redirect.com",
    ]



import cgi
from io import BytesIO
import pytest
from nylas.client.errors import FileUploadError


@pytest.mark.usefixtures("mock_files")
def test_file_upload_data(api_client, mocked_responses):
    data = "Hello, World!"

    myfile = api_client.files.create()
    myfile.filename = "hello.txt"
    myfile.data = data

    assert not mocked_responses.calls
    myfile.save()
    assert len(mocked_responses.calls) == 1

    assert myfile.filename == "hello.txt"
    assert myfile.size == 13

    upload_body = mocked_responses.calls[0].request.body
    upload_lines = upload_body.decode("utf8").splitlines()

    content_disposition = upload_lines[1]
    _, params = cgi.parse_header(content_disposition)
    assert params["filename"] == "hello.txt"
    assert "Hello, World!" in upload_lines


@pytest.mark.usefixtures("mock_files")
def test_file_upload_stream(api_client, mocked_responses):
    stream = BytesIO(b"Hello, World!")
    stream.name = "wacky.txt"

    myfile = api_client.files.create()
    myfile.filename = "hello.txt"
    myfile.stream = stream
    assert not mocked_responses.calls
    myfile.save()
    assert len(mocked_responses.calls) == 1

    assert myfile.filename == "hello.txt"
    assert myfile.size == 13

    upload_body = mocked_responses.calls[0].request.body
    upload_lines = upload_body.decode("utf8").splitlines()

    content_disposition = upload_lines[1]
    _, params = cgi.parse_header(content_disposition)
    assert params["filename"] == "hello.txt"
    assert "Hello, World!" in upload_lines


@pytest.mark.usefixtures("mock_files")
def test_file_download(api_client, mocked_responses):
    assert not mocked_responses.calls
    myfile = api_client.files.first()
    assert len(mocked_responses.calls) == 1
    data = myfile.download().decode()
    assert len(mocked_responses.calls) == 2
    assert data == "Hello, World!"


def test_file_invalid_upload(api_client):
    myfile = api_client.files.create()
    with pytest.raises(FileUploadError) as exc:
        myfile.save()

    assert str(exc.value) == (
        "File object not properly formatted, " "must provide either a stream or data."
    )


def test_file_upload_errors(api_client):
    myfile = api_client.files.create()
    myfile.filename = "test.txt"
    myfile.data = "Hello World."

    with pytest.raises(FileUploadError) as exc:
        myfile.download()

    assert str(exc.value) == ("Can't download a file that " "hasn't been uploaded.")



import json
import random

import pytest
import responses
from urlobject import URLObject


def test_no_filter(mocked_responses, api_client, api_url, message_body):
    message_body_list_50 = [message_body for _ in range(1, 51)]
    message_body_list_22 = [message_body for _ in range(1, 23)]

    values = [
        (200, {}, json.dumps(message_body_list_22)),
        (200, {}, json.dumps(message_body_list_50)),
    ]

    def callback(_request):
        return values.pop()

    mocked_responses.add_callback(responses.GET, api_url + "/events", callback=callback)

    events = api_client.events.all()
    assert len(events) == 72
    assert events[0].id == "cv4ei7syx10uvsxbs21ccsezf"


def test_two_filters(mocked_responses, api_client, api_url):
    mocked_responses.add(responses.GET, api_url + "/events", body="[]")
    events = api_client.events.where(param1="a", param2="b").all()
    assert len(events) == 0  # pylint: disable=len-as-condition
    url = mocked_responses.calls[-1].request.url
    query = URLObject(url).query_dict
    assert query["param1"] == "a"
    assert query["param2"] == "b"


@pytest.mark.usefixtures("mock_event_create_response_with_limits")
def test_limit_filter(mocked_responses, api_client, api_url, message_body):
    events = api_client.events.where(limit=51).all()
    assert len(events) == 51  # pylint: disable=len-as-condition
    url = mocked_responses.calls[-1].request.url
    query = URLObject(url).query_dict
    assert query["limit"] == "51"


def test_no_offset(mocked_responses, api_client, api_url):
    mocked_responses.add(responses.GET, api_url + r"/events", body="[]")
    list(api_client.events.where({"in": "Nylas"}).values())
    url = mocked_responses.calls[-1].request.url
    query = URLObject(url).query_dict
    assert query["in"] == "Nylas"
    assert query["offset"] == "0"


def test_zero_offset(mocked_responses, api_client, api_url):
    mocked_responses.add(responses.GET, api_url + "/events", body="[]")
    list(api_client.events.where({"in": "Nylas", "offset": 0}).values())
    url = mocked_responses.calls[-1].request.url
    query = URLObject(url).query_dict
    assert query["in"] == "Nylas"
    assert query["offset"] == "0"


def test_non_zero_offset(mocked_responses, api_client, api_url):
    offset = random.randint(1, 1000)
    mocked_responses.add(responses.GET, api_url + "/events", body="[]")

    list(api_client.events.where({"in": "Nylas", "offset": offset}).values())
    url = mocked_responses.calls[-1].request.url
    query = URLObject(url).query_dict
    assert query["in"] == "Nylas"
    assert query["offset"] == str(offset)



import json

from requests import HTTPError


class NylasError(Exception):
    pass


class MessageRejectedError(NylasError):
    pass


class FileUploadError(NylasError):
    pass


class UnSyncedError(NylasError):
    """
    HTTP Code 202
    The request was valid but the resource wasn't ready. Retry the request with exponential backoff.
    """

    pass


class NylasApiError(HTTPError):
    """
    Error class for Nylas API Errors
    This class provides more information to the user sent from the server, if present
    """

    def __init__(self, response):
        try:
            response_json = json.loads(response.text)
            error_message = "%s %s. Reason: %s. Nylas Error Type: %s" % (
                response.status_code,
                response.reason,
                response_json["message"],
                response_json["type"],
            )
            super(NylasApiError, self).__init__(error_message, response=response)
        except (ValueError, KeyError):
            super(NylasApiError, self).__init__(response.text, response=response)


class RateLimitError(NylasApiError):
    """
    Error class for 429 rate limit errors
    This class provides details about the rate limit returned from the server
    """

    RATE_LIMIT_LIMIT_HEADER = "X-RateLimit-Limit"
    RATE_LIMIT_RESET_HEADER = "X-RateLimit-Reset"

    def __init__(self, response):
        try:
            self.rate_limit = int(response.headers[self.RATE_LIMIT_LIMIT_HEADER])
            self.rate_limit_reset = int(response.headers[self.RATE_LIMIT_RESET_HEADER])
            super(RateLimitError, self).__init__(response)
        except (ValueError, KeyError):
            super(RateLimitError, self).__init__(response)



import json
import re
import pytest
import responses
import six
from requests import RequestException
from nylas.client.errors import MessageRejectedError, RateLimitError


def mock_sending_error(
    http_code, message, mocked_responses, api_url, server_error=None, headers=None
):
    send_endpoint = re.compile(api_url + "/send")
    response_body = {"type": "api_error", "message": message}

    if six.PY2 and http_code == 429:
        # Python 2 `httplib` doesn't know about status code 429
        six.moves.http_client.responses[429] = "Too Many Requests"

    if server_error is not None:
        response_body["server_error"] = server_error

    response_body = json.dumps(response_body)
    mocked_responses.add(
        responses.POST,
        send_endpoint,
        content_type="application/json",
        status=http_code,
        body=response_body,
        headers=headers,
    )


@pytest.mark.usefixtures("mock_account", "mock_save_draft")
def test_handle_message_rejected(mocked_responses, api_client, api_url):
    draft = api_client.drafts.create()
    error_message = "Sending to all recipients failed"
    mock_sending_error(402, error_message, mocked_responses, api_url=api_url)
    with pytest.raises(MessageRejectedError):
        draft.send()


@pytest.mark.usefixtures("mock_account", "mock_save_draft")
def test_handle_quota_exceeded(mocked_responses, api_client, api_url):
    draft = api_client.drafts.create()
    error_message = "Daily sending quota exceeded"
    error_headers = {"X-RateLimit-Limit": "500", "X-RateLimit-Reset": "10"}
    mock_sending_error(
        429, error_message, mocked_responses, api_url=api_url, headers=error_headers
    )
    with pytest.raises(RateLimitError) as exc:
        draft.send()
    assert "Too Many Requests" in str(exc.value)
    assert exc.value.rate_limit == 500
    assert exc.value.rate_limit_reset == 10


@pytest.mark.usefixtures("mock_account", "mock_save_draft")
def test_handle_quota_exceeded_no_headers(mocked_responses, api_client, api_url):
    draft = api_client.drafts.create()
    error_message = "Daily sending quota exceeded"
    mock_sending_error(429, error_message, mocked_responses, api_url=api_url)
    with pytest.raises(RateLimitError) as exc:
        draft.send()
    assert "Too Many Requests" in str(exc.value)


@pytest.mark.usefixtures("mock_account", "mock_save_draft")
def test_handle_service_unavailable(mocked_responses, api_client, api_url):
    draft = api_client.drafts.create()
    error_message = "The server unexpectedly closed the connection"
    mock_sending_error(503, error_message, mocked_responses, api_url=api_url)
    with pytest.raises(RequestException) as exc:
        draft.send()
    assert "Service Unavailable" in str(exc.value)


@pytest.mark.usefixtures("mock_account", "mock_save_draft")
def test_returns_server_error(mocked_responses, api_client, api_url):
    draft = api_client.drafts.create()
    error_message = "The server unexpectedly closed the connection"
    reason = "Rejected potential SPAM"
    mock_sending_error(
        503, error_message, mocked_responses, api_url=api_url, server_error=reason
    )
    with pytest.raises(RequestException) as exc:
        draft.send()

    assert "Service Unavailable" in str(exc.value)


@pytest.mark.usefixtures("mock_account", "mock_save_draft")
def test_doesnt_return_server_error_if_not_defined(
    mocked_responses, api_client, api_url
):
    draft = api_client.drafts.create()
    error_message = "The server unexpectedly closed the connection"
    mock_sending_error(503, error_message, mocked_responses, api_url=api_url)
    with pytest.raises(RequestException) as exc:
        draft.send()
    assert "Service Unavailable" in str(exc.value)



import json
from datetime import datetime, timedelta
import pytest
from urlobject import URLObject
from nylas.client.restful_models import Event


def blank_event(api_client):
    event = api_client.events.create()
    event.title = "Paris-Brest"
    event.calendar_id = "calendar_id"
    event.when = {"start_time": 1409594400, "end_time": 1409594400}
    return event


@pytest.mark.usefixtures("mock_event_create_response")
def test_event_crud(mocked_responses, api_client):
    event1 = blank_event(api_client)
    event1.object = "should not send"
    event1.account_id = "should not send"
    event1.job_status_id = "should not send"
    event1.ical_uid = "should not send"
    event1.message_id = "should not send"
    event1.owner = "should not send"
    event1.status = "should not send"
    event1.master_event_id = "should not send"
    event1.original_start_time = "should not send"
    event1.visibility = "private"
    event1.save()
    request = mocked_responses.calls[0].request
    body = json.loads(request.body)
    assert event1.id == "cv4ei7syx10uvsxbs21ccsezf"
    assert event1.visibility == "private"
    assert "title" in body
    assert "object" not in body
    assert "account_id" not in body
    assert "job_status_id" not in body
    assert "ical_uid" not in body
    assert "message_id" not in body
    assert "owner" not in body
    assert "status" not in body
    assert "master_event_id" not in body
    assert "original_start_time" not in body

    event1.title = "blah"
    event1.save()
    request = mocked_responses.calls[1].request
    body = json.loads(request.body)
    assert event1.title == "loaded from JSON"
    assert event1.get("ignored") is None
    assert "id" not in body


@pytest.mark.usefixtures("mock_event_create_notify_response")
def test_event_notify(mocked_responses, api_client):
    event1 = blank_event(api_client)
    event1.save(notify_participants="true", other_param="1")
    assert event1.id == "cv4ei7syx10uvsxbs21ccsezf"

    url = mocked_responses.calls[-1].request.url
    query = URLObject(url).query_dict
    assert query["notify_participants"] == "true"
    assert query["other_param"] == "1"


@pytest.mark.usefixtures("mock_event_create_response")
def test_event_conferencing_details(mocked_responses, api_client):
    event = blank_event(api_client)
    event.conferencing = {
        "provider": "Zoom Meeting",
        "details": {
            "url": "https://us02web.zoom.us/j/****************",
            "meeting_code": "213",
            "password": "xyz",
            "phone": ["+11234567890"],
        },
    }
    event.save()
    assert event.id == "cv4ei7syx10uvsxbs21ccsezf"
    assert event.conferencing["provider"] == "Zoom Meeting"
    assert (
        event.conferencing["details"]["url"]
        == "https://us02web.zoom.us/j/****************"
    )
    assert event.conferencing["details"]["meeting_code"] == "213"
    assert event.conferencing["details"]["password"] == "xyz"
    assert event.conferencing["details"]["phone"] == ["+11234567890"]

    body = json.loads(mocked_responses.calls[-1].request.body)
    assert body["conferencing"]["provider"] == "Zoom Meeting"
    assert (
        body["conferencing"]["details"]["url"]
        == "https://us02web.zoom.us/j/****************"
    )
    assert body["conferencing"]["details"]["meeting_code"] == "213"
    assert body["conferencing"]["details"]["password"] == "xyz"
    assert body["conferencing"]["details"]["phone"] == ["+11234567890"]


@pytest.mark.usefixtures("mock_event_create_response")
def test_event_conferencing_autocreate(mocked_responses, api_client):
    event = blank_event(api_client)
    event.conferencing = {
        "provider": "Zoom Meeting",
        "autocreate": {
            "settings": {},
        },
    }
    event.save()
    assert event.id == "cv4ei7syx10uvsxbs21ccsezf"
    assert event.conferencing["provider"] == "Zoom Meeting"
    assert event.conferencing["autocreate"]["settings"] == {}

    body = json.loads(mocked_responses.calls[-1].request.body)
    assert body["conferencing"]["provider"] == "Zoom Meeting"
    assert event.conferencing["autocreate"]["settings"] == {}


@pytest.mark.usefixtures("mock_event_create_response")
def test_event_conferencing_details_autocreate_error(mocked_responses, api_client):
    event = blank_event(api_client)
    event.conferencing = {
        "provider": "Zoom Meeting",
        "details": {
            "url": "https://us02web.zoom.us/j/****************",
            "meeting_code": "213",
            "password": "xyz",
            "phone": ["+11234567890"],
        },
        "autocreate": {
            "settings": {
                "password": "1234",
            },
        },
    }
    with pytest.raises(ValueError) as excinfo:
        event.save()
    assert "Cannot set both 'details' and 'autocreate' in conferencing object." in str(
        excinfo
    )


@pytest.mark.usefixtures("mock_event_create_response")
def test_event_error_if_participants_more_than_capacity(mocked_responses, api_client):
    event = blank_event(api_client)
    event.capacity = 1
    event.participants = [
        {"email": "person1@email.com"},
        {"email": "person2@email.com"},
    ]
    with pytest.raises(ValueError) as excinfo:
        event.save()
    assert "The number of participants in the event exceeds the set capacity." in str(
        excinfo
    )


@pytest.mark.usefixtures("mock_event_create_response")
def test_event_no_error_if_capacity_negative_one(mocked_responses, api_client):
    event = blank_event(api_client)
    event.capacity = -1
    event.participants = [
        {"email": "person1@email.com"},
        {"email": "person2@email.com"},
    ]
    event.save()


@pytest.mark.usefixtures("mock_event_create_response")
def test_event_no_error_if_participants_less_than_eql_capacity(
    mocked_responses, api_client
):
    event = blank_event(api_client)
    event.capacity = 2
    event.participants = [
        {"email": "person1@email.com"},
        {"email": "person2@email.com"},
    ]
    event.save()
    event.capacity = 3
    event.participants = [
        {"email": "person1@email.com"},
        {"email": "person2@email.com"},
    ]
    event.save()


@pytest.mark.usefixtures("mock_calendars", "mock_events")
def test_calendar_events(api_client):
    calendar = api_client.calendars.first()
    assert calendar.events
    assert all(isinstance(event, Event) for event in calendar.events)


@pytest.mark.usefixtures("mock_events", "mock_send_rsvp")
def test_event(mocked_responses, api_client):
    event = api_client.events.first()
    event.rsvp("yes")

    request = mocked_responses.calls[-1].request
    assert URLObject(request.url).path == "/send-rsvp"
    data = json.loads(request.body)
    assert data["event_id"] == event.id
    assert data["status"] == "yes"
    assert data["comment"] == None


@pytest.mark.usefixtures("mock_events", "mock_send_rsvp")
def test_event_rsvp_with_comment(mocked_responses, api_client):
    event = api_client.events.first()
    event.rsvp("no", "I have a conflict")

    request = mocked_responses.calls[-1].request
    assert URLObject(request.url).path == "/send-rsvp"
    data = json.loads(request.body)
    assert data["event_id"] == event.id
    assert data["status"] == "no"
    assert data["comment"] == "I have a conflict"


@pytest.mark.usefixtures("mock_events")
def test_event_rsvp_invalid(api_client):
    event = api_client.events.first()
    with pytest.raises(ValueError) as excinfo:
        event.rsvp("purple")
    assert "invalid status" in str(excinfo)


@pytest.mark.usefixtures("mock_events")
def test_event_rsvp_no_message(api_client):
    event = api_client.events.all()[1]
    with pytest.raises(ValueError) as excinfo:
        event.rsvp("yes")
    assert "This event was not imported from an iCalendar invite" in str(excinfo)


@pytest.mark.usefixtures("mock_free_busy")
def test_free_busy_datetime(mocked_responses, api_client):
    email = "fake@example.com"
    start_at = datetime(2020, 1, 1)
    end_at = datetime(2020, 1, 2)
    free_busy = api_client.free_busy([email], start_at, end_at)

    assert isinstance(free_busy, list)
    assert isinstance(free_busy[0], dict)
    assert free_busy[0]["email"] == "fake@example.com"
    assert "time_slots" in free_busy[0]

    request = mocked_responses.calls[-1].request
    assert URLObject(request.url).path == "/calendars/free-busy"
    data = json.loads(request.body)
    assert data["emails"] == [email]
    assert data["start_time"] == 1577836800
    assert data["end_time"] == 1577923200


@pytest.mark.usefixtures("mock_free_busy")
def test_free_busy_timestamp(mocked_responses, api_client):
    email = "ron@example.com"
    start_time = 1580511600
    end_time = 1580598000
    free_busy = api_client.free_busy([email], start_time, end_time)

    assert isinstance(free_busy, list)
    assert isinstance(free_busy[0], dict)
    assert free_busy[0]["email"] == "ron@example.com"
    assert "time_slots" in free_busy[0]

    request = mocked_responses.calls[-1].request
    assert URLObject(request.url).path == "/calendars/free-busy"
    data = json.loads(request.body)
    assert data["emails"] == [email]
    assert data["start_time"] == 1580511600
    assert data["end_time"] == 1580598000


@pytest.mark.usefixtures("mock_free_busy")
def test_free_busy_single_email(mocked_responses, api_client):
    email = "ben@bitdiddle.com"
    start_at = datetime(2000, 1, 1)
    end_at = datetime(2000, 3, 1)
    free_busy = api_client.free_busy(email, start_at, end_at)

    assert isinstance(free_busy, list)
    assert isinstance(free_busy[0], dict)
    assert free_busy[0]["email"] == "ben@bitdiddle.com"
    assert "time_slots" in free_busy[0]

    request = mocked_responses.calls[-1].request
    assert URLObject(request.url).path == "/calendars/free-busy"
    data = json.loads(request.body)
    assert data["emails"] == [email]
    assert data["start_time"] == 946684800
    assert data["end_time"] == 951868800


@pytest.mark.usefixtures("mock_free_busy")
def test_free_busy_with_calendars(mocked_responses, api_client):
    email = "ben@bitdiddle.com"
    start_at = datetime(2000, 1, 1)
    end_at = datetime(2000, 3, 1)
    calendars = [
        {
            "account_id": "test_account_id",
            "calendar_ids": ["example_calendar_a", "example_calendar_b"],
        }
    ]
    api_client.free_busy([email], start_at, end_at, calendars)

    request = mocked_responses.calls[-1].request
    assert URLObject(request.url).path == "/calendars/free-busy"
    data = json.loads(request.body)
    assert data["emails"] == [email]
    assert data["start_time"] == 946684800
    assert data["end_time"] == 951868800
    assert len(data["calendars"]) == 1
    assert data["calendars"][0] == {
        "account_id": "test_account_id",
        "calendar_ids": ["example_calendar_a", "example_calendar_b"],
    }


@pytest.mark.usefixtures("mock_free_busy")
def test_free_busy_without_emails_or_calendar(mocked_responses, api_client):
    start_at = datetime(2000, 1, 1)
    end_at = datetime(2000, 3, 1)
    with pytest.raises(ValueError) as excinfo:
        api_client.free_busy([], start_at, end_at)
    assert "Must set either 'emails' or 'calendars' in the query." in str(excinfo)


@pytest.mark.usefixtures("mock_availability")
def test_availability_datetime(mocked_responses, api_client):
    emails = ["one@example.com", "two@example.com", "three@example.com"]
    duration = timedelta(minutes=30)
    interval = timedelta(hours=1, minutes=30)
    start_at = datetime(2020, 1, 1)
    end_at = datetime(2020, 1, 2)
    availability = api_client.availability(emails, duration, interval, start_at, end_at)

    assert isinstance(availability, dict)
    assert "time_slots" in availability

    request = mocked_responses.calls[-1].request
    assert URLObject(request.url).path == "/calendars/availability"
    data = json.loads(request.body)
    assert data["emails"] == emails
    assert data["duration_minutes"] == 30
    assert isinstance(data["duration_minutes"], int)
    assert data["interval_minutes"] == 90
    assert isinstance(data["interval_minutes"], int)
    assert data["start_time"] == 1577836800
    assert data["end_time"] == 1577923200
    assert data["free_busy"] == []


@pytest.mark.usefixtures("mock_availability")
def test_availability_timestamp(mocked_responses, api_client):
    emails = ["one@example.com", "two@example.com", "three@example.com"]
    duration = 30
    interval = 60
    start_time = 1580511600
    end_time = 1580598000
    availability = api_client.availability(
        emails, duration, interval, start_time, end_time
    )

    assert isinstance(availability, dict)
    assert "time_slots" in availability

    request = mocked_responses.calls[-1].request
    assert URLObject(request.url).path == "/calendars/availability"
    data = json.loads(request.body)
    assert data["emails"] == emails
    assert data["duration_minutes"] == 30
    assert isinstance(data["duration_minutes"], int)
    assert data["interval_minutes"] == 60
    assert isinstance(data["interval_minutes"], int)
    assert data["start_time"] == 1580511600
    assert data["end_time"] == 1580598000
    assert data["free_busy"] == []


@pytest.mark.usefixtures("mock_availability")
def test_availability_single_email(mocked_responses, api_client):
    email = "ben@bitdiddle.com"
    duration = timedelta(minutes=60)
    interval = 5
    start_at = datetime(2000, 1, 1)
    end_at = datetime(2000, 3, 1)
    availability = api_client.availability(email, duration, interval, start_at, end_at)

    assert isinstance(availability, dict)
    assert "time_slots" in availability

    request = mocked_responses.calls[-1].request
    assert URLObject(request.url).path == "/calendars/availability"
    data = json.loads(request.body)
    assert data["emails"] == [email]
    assert data["duration_minutes"] == 60
    assert isinstance(data["duration_minutes"], int)
    assert data["interval_minutes"] == 5
    assert isinstance(data["interval_minutes"], int)
    assert data["start_time"] == 946684800
    assert data["end_time"] == 951868800
    assert data["free_busy"] == []


@pytest.mark.usefixtures("mock_availability")
def test_availability_with_free_busy(mocked_responses, api_client):
    emails = [
        "one@example.com",
        "two@example.com",
        "three@example.com",
        "visitor@external.net",
    ]
    duration = 48
    interval = timedelta(minutes=18)
    start_at = datetime(2020, 1, 1)
    end_at = datetime(2020, 1, 2)
    free_busy = [
        {
            "email": "visitor@external.net",
            "time_slots": [
                {
                    "object": "time_slot",
                    "status": "busy",
                    "start_time": 1584377898,
                    "end_time": 1584379800,
                }
            ],
            "object": "free_busy",
        }
    ]
    availability = api_client.availability(
        emails, duration, interval, start_at, end_at, free_busy=free_busy
    )

    assert isinstance(availability, dict)
    assert "time_slots" in availability

    request = mocked_responses.calls[-1].request
    assert URLObject(request.url).path == "/calendars/availability"
    data = json.loads(request.body)
    assert data["emails"] == emails
    assert data["duration_minutes"] == 48
    assert isinstance(data["duration_minutes"], int)
    assert data["interval_minutes"] == 18
    assert isinstance(data["interval_minutes"], int)
    assert data["start_time"] == 1577836800
    assert data["end_time"] == 1577923200
    assert data["free_busy"] == free_busy


@pytest.mark.usefixtures("mock_availability")
def test_availability_with_calendars(mocked_responses, api_client):
    emails = [
        "one@example.com",
        "two@example.com",
        "three@example.com",
        "visitor@external.net",
    ]
    duration = 48
    interval = timedelta(minutes=18)
    start_at = datetime(2020, 1, 1)
    end_at = datetime(2020, 1, 2)
    calendars = [
        {
            "account_id": "test_account_id",
            "calendar_ids": ["example_calendar_a", "example_calendar_b"],
        }
    ]
    api_client.availability(
        emails, duration, interval, start_at, end_at, calendars=calendars
    )

    request = mocked_responses.calls[-1].request
    assert URLObject(request.url).path == "/calendars/availability"
    data = json.loads(request.body)
    assert data["emails"] == emails
    assert data["duration_minutes"] == 48
    assert isinstance(data["duration_minutes"], int)
    assert data["interval_minutes"] == 18
    assert isinstance(data["interval_minutes"], int)
    assert data["start_time"] == 1577836800
    assert data["end_time"] == 1577923200
    assert len(data["calendars"]) == 1
    assert data["calendars"][0] == {
        "account_id": "test_account_id",
        "calendar_ids": ["example_calendar_a", "example_calendar_b"],
    }


@pytest.mark.usefixtures("mock_availability")
def test_availability_without_emails_or_calendar(mocked_responses, api_client):
    duration = 48
    interval = timedelta(minutes=18)
    start_at = datetime(2000, 1, 1)
    end_at = datetime(2000, 3, 1)
    with pytest.raises(ValueError) as excinfo:
        api_client.availability([], duration, interval, start_at, end_at)
    assert "Must set either 'emails' or 'calendars' in the query." in str(excinfo)


@pytest.mark.usefixtures("mock_availability")
def test_consecutive_availability(mocked_responses, api_client):
    emails = [["one@example.com"], ["two@example.com", "three@example.com"]]
    duration = timedelta(minutes=30)
    interval = timedelta(hours=1, minutes=30)
    start_at = datetime(2020, 1, 1)
    end_at = datetime(2020, 1, 2)
    open_hours = api_client.open_hours(
        ["one@example.com", "two@example.com", "three@example.com"],
        [0],
        "America/Chicago",
        "10:00",
        "14:00",
    )
    api_client.consecutive_availability(
        emails, duration, interval, start_at, end_at, open_hours=[open_hours]
    )

    request = mocked_responses.calls[-1].request
    assert URLObject(request.url).path == "/calendars/availability/consecutive"
    data = json.loads(request.body)
    assert data["emails"] == emails
    assert data["duration_minutes"] == 30
    assert isinstance(data["duration_minutes"], int)
    assert data["interval_minutes"] == 90
    assert isinstance(data["interval_minutes"], int)
    assert data["start_time"] == 1577836800
    assert data["end_time"] == 1577923200
    assert data["free_busy"] == []
    assert data["open_hours"][0]["emails"] == [
        "one@example.com",
        "two@example.com",
        "three@example.com",
    ]
    assert data["open_hours"][0]["days"] == [0]
    assert data["open_hours"][0]["timezone"] == "America/Chicago"
    assert data["open_hours"][0]["start"] == "10:00"
    assert data["open_hours"][0]["end"] == "14:00"


@pytest.mark.usefixtures("mock_availability")
def test_consecutive_availability_free_busy(mocked_responses, api_client):
    emails = [["one@example.com"], ["two@example.com", "three@example.com"]]
    duration = timedelta(minutes=30)
    interval = timedelta(hours=1, minutes=30)
    start_at = datetime(2020, 1, 1)
    end_at = datetime(2020, 1, 2)
    open_hours = api_client.open_hours(
        [
            "one@example.com",
            "two@example.com",
            "three@example.com",
            "visitor@external.net",
        ],
        [0],
        "America/Chicago",
        "10:00",
        "14:00",
    )
    free_busy = [
        {
            "email": "visitor@external.net",
            "time_slots": [
                {
                    "object": "time_slot",
                    "status": "busy",
                    "start_time": 1584377898,
                    "end_time": 1584379800,
                }
            ],
            "object": "free_busy",
        }
    ]
    api_client.consecutive_availability(
        emails,
        duration,
        interval,
        start_at,
        end_at,
        free_busy=free_busy,
        open_hours=[open_hours],
    )

    request = mocked_responses.calls[-1].request
    assert URLObject(request.url).path == "/calendars/availability/consecutive"
    data = json.loads(request.body)
    assert data["emails"] == emails
    assert data["duration_minutes"] == 30
    assert isinstance(data["duration_minutes"], int)
    assert data["interval_minutes"] == 90
    assert isinstance(data["interval_minutes"], int)
    assert data["start_time"] == 1577836800
    assert data["end_time"] == 1577923200
    assert data["free_busy"] == free_busy
    assert data["open_hours"][0]["emails"] == [
        "one@example.com",
        "two@example.com",
        "three@example.com",
        "visitor@external.net",
    ]
    assert data["open_hours"][0]["days"] == [0]
    assert data["open_hours"][0]["timezone"] == "America/Chicago"
    assert data["open_hours"][0]["start"] == "10:00"
    assert data["open_hours"][0]["end"] == "14:00"


@pytest.mark.usefixtures("mock_availability")
def test_consecutive_availability_with_calendars(mocked_responses, api_client):
    emails = [["one@example.com"], ["two@example.com", "three@example.com"]]
    duration = timedelta(minutes=30)
    interval = timedelta(hours=1, minutes=30)
    start_at = datetime(2020, 1, 1)
    end_at = datetime(2020, 1, 2)
    calendars = [
        {
            "account_id": "test_account_id",
            "calendar_ids": ["example_calendar_a", "example_calendar_b"],
        }
    ]
    api_client.consecutive_availability(
        emails, duration, interval, start_at, end_at, calendars=calendars
    )

    request = mocked_responses.calls[-1].request
    assert URLObject(request.url).path == "/calendars/availability/consecutive"
    data = json.loads(request.body)
    assert data["emails"] == emails
    assert data["duration_minutes"] == 30
    assert isinstance(data["duration_minutes"], int)
    assert data["interval_minutes"] == 90
    assert isinstance(data["interval_minutes"], int)
    assert data["start_time"] == 1577836800
    assert data["end_time"] == 1577923200
    assert len(data["calendars"]) == 1
    assert data["calendars"][0] == {
        "account_id": "test_account_id",
        "calendar_ids": ["example_calendar_a", "example_calendar_b"],
    }


@pytest.mark.usefixtures("mock_availability")
def test_availability_without_emails_or_calendar(mocked_responses, api_client):
    duration = 48
    interval = timedelta(minutes=18)
    start_at = datetime(2000, 1, 1)
    end_at = datetime(2000, 3, 1)
    with pytest.raises(ValueError) as excinfo:
        api_client.consecutive_availability([], duration, interval, start_at, end_at)
    assert "Must set either 'emails' or 'calendars' in the query." in str(excinfo)


@pytest.mark.usefixtures("mock_availability")
def test_consecutive_availability_invalid_open_hours_email(
    mocked_responses, api_client
):
    emails = [["one@example.com"], ["two@example.com", "three@example.com"]]
    duration = timedelta(minutes=30)
    interval = timedelta(hours=1, minutes=30)
    start_at = datetime(2020, 1, 1)
    end_at = datetime(2020, 1, 2)
    open_hours = api_client.open_hours(
        [
            "one@example.com",
            "two@example.com",
            "three@example.com",
            "visitor@external.net",
            "four@example.com",
        ],
        [0],
        "America/Chicago",
        "10:00",
        "14:00",
    )
    free_busy = [
        {
            "email": "visitor@external.net",
            "time_slots": [
                {
                    "object": "time_slot",
                    "status": "busy",
                    "start_time": 1584377898,
                    "end_time": 1584379800,
                }
            ],
            "object": "free_busy",
        }
    ]
    with pytest.raises(ValueError):
        api_client.consecutive_availability(
            emails,
            duration,
            interval,
            start_at,
            end_at,
            free_busy=free_busy,
            open_hours=[open_hours],
        )


@pytest.mark.usefixtures("mock_events")
def test_metadata_filtering(api_client):
    events_filtered_by_key = api_client.events.where(metadata_key="platform")
    assert len(events_filtered_by_key.all()) > 0
    for event in events_filtered_by_key:
        assert "platform" in event["metadata"]

    events_filtered_by_value = api_client.events.where(
        metadata_value=["meeting", "java"]
    )
    assert len(events_filtered_by_value.all()) > 0
    for event in events_filtered_by_value:
        assert event["metadata"]["event_type"] == "meeting"

    events_filtered_by_pair = api_client.events.where(
        metadata_pair={"platform": "python", "bla": "blablabla"}
    )
    assert len(events_filtered_by_pair.all()) > 0
    for event in events_filtered_by_pair:
        assert "platform" in event["metadata"]
        assert event["metadata"]["platform"] == "python"

    non_existant_event = api_client.events.where(metadata_pair={"bla": "blablabla"})
    assert len(non_existant_event.all()) == 0


@pytest.mark.usefixtures("mock_event_create_response")
def test_event_notifications(mocked_responses, api_client):
    event = blank_event(api_client)
    event.notifications = [
        {
            "type": "email",
            "minutes_before_event": 60,
            "subject": "Test Event Notification",
            "body": "Reminding you about our meeting.",
        }
    ]
    event.save()
    assert event.id == "cv4ei7syx10uvsxbs21ccsezf"
    assert len(event.notifications) == 1
    assert event.notifications[0]["type"] == "email"
    assert event.notifications[0]["minutes_before_event"] == 60
    assert event.notifications[0]["subject"] == "Test Event Notification"
    assert event.notifications[0]["body"] == "Reminding you about our meeting."


@pytest.mark.usefixtures("mock_event_create_response", "mock_event_generate_ics")
def test_generate_ics_existing_event(mocked_responses, api_client):
    event = blank_event(api_client)
    event.save()
    ics = event.generate_ics()
    ics_request = mocked_responses.calls[1].request
    assert len(mocked_responses.calls) == 2
    assert event.id == "cv4ei7syx10uvsxbs21ccsezf"
    assert ics_request.path_url == "/events/to-ics"
    assert ics_request.method == "POST"
    assert json.loads(ics_request.body) == {"event_id": "cv4ei7syx10uvsxbs21ccsezf"}


@pytest.mark.usefixtures("mock_event_create_response", "mock_event_generate_ics")
def test_generate_ics_no_event_id(mocked_responses, api_client):
    event = blank_event(api_client)
    ics = event.generate_ics()
    ics_request = mocked_responses.calls[0].request
    assert len(mocked_responses.calls) == 1
    assert event.id is None
    assert ics_request.path_url == "/events/to-ics"
    assert ics_request.method == "POST"
    assert json.loads(ics_request.body) == {
        "calendar_id": "calendar_id",
        "title": "Paris-Brest",
        "when": {"end_time": 1409594400, "start_time": 1409594400},
    }


@pytest.mark.usefixtures("mock_event_create_response", "mock_event_generate_ics")
def test_generate_ics_options(mocked_responses, api_client):
    event = blank_event(api_client)
    event.save()
    ics = event.generate_ics(
        ical_uid="test_uuid", method="request", prodid="test_prodid"
    )
    ics_request = mocked_responses.calls[1].request
    assert len(mocked_responses.calls) == 2
    assert event.id == "cv4ei7syx10uvsxbs21ccsezf"
    assert ics_request.path_url == "/events/to-ics"
    assert ics_request.method == "POST"
    assert json.loads(ics_request.body) == {
        "event_id": "cv4ei7syx10uvsxbs21ccsezf",
        "ics_options": {
            "ical_uid": "test_uuid",
            "method": "request",
            "prodid": "test_prodid",
        },
    }


@pytest.mark.usefixtures("mock_event_create_response", "mock_event_generate_ics")
def test_generate_ics_no_calendar_id_throws(mocked_responses, api_client):
    event = blank_event(api_client)
    del event.calendar_id
    with pytest.raises(ValueError) as exc:
        event.generate_ics()

    assert str(exc.value) == (
        "Cannot generate an ICS file for an event without a Calendar ID or when set"
    )


@pytest.mark.usefixtures("mock_event_create_response", "mock_event_generate_ics")
def test_generate_ics_no_when_throws(mocked_responses, api_client):
    event = blank_event(api_client)
    del event.when
    with pytest.raises(ValueError) as exc:
        event.generate_ics()

    assert str(exc.value) == (
        "Cannot generate an ICS file for an event without a Calendar ID or when set"
    )



import json
from datetime import datetime
import pytest
import responses

from nylas.client.restful_models import Scheduler, Calendar
from nylas.client.scheduler_models import SchedulerTimeSlot, SchedulerBookingRequest


def blank_scheduler_page(api_client):
    scheduler = api_client.scheduler.create()
    scheduler.access_tokens = ["test-access-token"]
    scheduler.name = "Python SDK Example"
    scheduler.slug = "py_example_1"
    return scheduler


def test_scheduler_endpoint(api_client):
    scheduler = api_client.scheduler
    assert scheduler.api.api_server == "https://api.schedule.nylas.com"


@pytest.mark.usefixtures("mock_schedulers")
def test_scheduler(api_client):
    scheduler = api_client.scheduler.first()
    assert isinstance(scheduler, Scheduler)
    assert scheduler.id == 90210
    assert scheduler.app_client_id == "test-client-id"
    assert scheduler.app_organization_id == 12345
    assert len(scheduler.config) == 4
    assert isinstance(scheduler.config, dict)
    assert scheduler.config["locale"] == "en"
    assert len(scheduler.config["reminders"]) == 0
    assert scheduler.config["timezone"] == "America/Los_Angeles"
    assert scheduler.edit_token == "test-edit-token-1"
    assert scheduler.name == "test-1"
    assert scheduler.slug == "test1"
    assert scheduler.created_at == datetime.strptime("2021-10-22", "%Y-%m-%d").date()
    assert scheduler.modified_at == datetime.strptime("2021-10-22", "%Y-%m-%d").date()


@pytest.mark.usefixtures("mock_scheduler_create_response")
def test_create_scheduler(api_client):
    scheduler = blank_scheduler_page(api_client)
    scheduler.save()
    assert scheduler.id == "cv4ei7syx10uvsxbs21ccsezf"


@pytest.mark.usefixtures("mock_scheduler_create_response")
def test_modify_scheduler(api_client):
    scheduler = blank_scheduler_page(api_client)
    scheduler.id = "cv4ei7syx10uvsxbs21ccsezf"
    scheduler.name = "Updated Name"
    scheduler.save()
    assert scheduler.name == "Updated Name"


@pytest.mark.usefixtures("mock_scheduler_get_available_calendars")
def test_scheduler_get_available_calendars(api_client):
    scheduler = blank_scheduler_page(api_client)
    scheduler.id = "cv4ei7syx10uvsxbs21ccsezf"
    calendars = scheduler.get_available_calendars()
    assert len(calendars) == 1
    calendar = calendars[0]
    assert len(calendar["calendars"]) == 1
    assert isinstance(calendar["calendars"][0], Calendar)
    assert calendar["calendars"][0].id == "calendar-id"
    assert calendar["calendars"][0].name == "Emailed events"
    assert calendar["calendars"][0].read_only
    assert calendar["email"] == "swag@nylas.com"
    assert calendar["id"] == "scheduler-id"
    assert calendar["name"] == "Python Tester"


@pytest.mark.usefixtures("mock_scheduler_get_available_calendars")
def test_scheduler_get_available_calendars_no_id_throws_error(api_client):
    scheduler = blank_scheduler_page(api_client)
    with pytest.raises(ValueError):
        scheduler.get_available_calendars()


@pytest.mark.usefixtures("mock_scheduler_upload_image")
def test_scheduler_upload_image(api_client):
    scheduler = blank_scheduler_page(api_client)
    scheduler.id = "cv4ei7syx10uvsxbs21ccsezf"
    upload = scheduler.upload_image("image/png", "test.png")
    assert upload["filename"] == "test.png"
    assert upload["originalFilename"] == "test.png"
    assert upload["publicUrl"] == "https://public.nylas.com/test.png"
    assert upload["signedUrl"] == "https://signed.nylas.com/test.png"


@pytest.mark.usefixtures("mock_scheduler_get_available_calendars")
def test_scheduler_get_available_calendars_no_id_throws_error(api_client):
    scheduler = blank_scheduler_page(api_client)
    with pytest.raises(ValueError):
        scheduler.upload_image("image/png", "test.png")


@pytest.mark.usefixtures("mock_scheduler_provider_availability")
def test_scheduler_get_google_availability(mocked_responses, api_client):
    api_client.scheduler.get_google_availability()
    request = mocked_responses.calls[0].request
    assert request.url == "https://api.schedule.nylas.com/schedule/availability/google"
    assert request.method == responses.GET


@pytest.mark.usefixtures("mock_scheduler_provider_availability")
def test_scheduler_get_o365_availability(mocked_responses, api_client):
    api_client.scheduler.get_office_365_availability()
    request = mocked_responses.calls[0].request
    assert request.url == "https://api.schedule.nylas.com/schedule/availability/o365"
    assert request.method == responses.GET


@pytest.mark.usefixtures("mock_schedulers")
def test_scheduler_get_page_slug(mocked_responses, api_client):
    scheduler = api_client.scheduler.get_page_slug("test1")
    request = mocked_responses.calls[0].request
    assert request.url == "https://api.schedule.nylas.com/schedule/test1/info"
    assert request.method == responses.GET
    assert isinstance(scheduler, Scheduler)
    assert scheduler.id == 90210
    assert scheduler.app_client_id == "test-client-id"
    assert scheduler.app_organization_id == 12345
    assert len(scheduler.config) == 4
    assert isinstance(scheduler.config, dict)
    assert scheduler.config["locale"] == "en"
    assert len(scheduler.config["reminders"]) == 0
    assert scheduler.config["timezone"] == "America/Los_Angeles"
    assert scheduler.edit_token == "test-edit-token-1"
    assert scheduler.name == "test-1"
    assert scheduler.slug == "test1"
    assert scheduler.created_at == datetime.strptime("2021-10-22", "%Y-%m-%d").date()
    assert scheduler.modified_at == datetime.strptime("2021-10-22", "%Y-%m-%d").date()


@pytest.mark.usefixtures("mock_scheduler_timeslots")
def test_scheduler_get_available_time_slots(mocked_responses, api_client):
    scheduler = blank_scheduler_page(api_client)
    timeslots = api_client.scheduler.get_available_time_slots(scheduler.slug)
    request = mocked_responses.calls[0].request
    assert (
        request.url == "https://api.schedule.nylas.com/schedule/py_example_1/timeslots"
    )
    assert request.method == responses.GET
    assert len(timeslots) == 1
    assert timeslots[0]
    assert timeslots[0].account_id == "test-account-id"
    assert timeslots[0].calendar_id == "test-calendar-id"
    assert timeslots[0].emails[0] == "test@example.com"
    assert timeslots[0].host_name == "www.hostname.com"
    assert timeslots[0].end == datetime.utcfromtimestamp(1636731958)
    assert timeslots[0].start == datetime.utcfromtimestamp(1636728347)


@pytest.mark.usefixtures("mock_scheduler_timeslots")
def test_scheduler_get_available_time_slots(mocked_responses, api_client):
    scheduler = blank_scheduler_page(api_client)
    timeslots = api_client.scheduler.get_available_time_slots(scheduler.slug)
    request = mocked_responses.calls[0].request
    assert (
        request.url == "https://api.schedule.nylas.com/schedule/py_example_1/timeslots"
    )
    assert request.method == responses.GET
    assert len(timeslots) == 1
    assert timeslots[0]
    assert timeslots[0].account_id == "test-account-id"
    assert timeslots[0].calendar_id == "test-calendar-id"
    assert timeslots[0].emails[0] == "test@example.com"
    assert timeslots[0].host_name == "www.hostname.com"
    assert timeslots[0].end == datetime.utcfromtimestamp(1636731958)
    assert timeslots[0].start == datetime.utcfromtimestamp(1636728347)


@pytest.mark.usefixtures("mock_scheduler_timeslots")
def test_scheduler_book_time_slot(mocked_responses, api_client):
    scheduler = blank_scheduler_page(api_client)
    slot = SchedulerTimeSlot.create(api_client)
    slot.account_id = "test-account-id"
    slot.calendar_id = "test-calendar-id"
    slot.emails = ["recipient@example.com"]
    slot.host_name = "www.nylas.com"
    slot.start = datetime.utcfromtimestamp(1636728347)
    slot.end = datetime.utcfromtimestamp(1636731958)
    timeslot_to_book = SchedulerBookingRequest.create(api_client)
    timeslot_to_book.additional_values = {
        "test": "yes",
    }
    timeslot_to_book.email = "recipient@example.com"
    timeslot_to_book.locale = "en_US"
    timeslot_to_book.name = "Recipient Doe"
    timeslot_to_book.timezone = "America/New_York"
    timeslot_to_book.slot = slot
    booking_response = api_client.scheduler.book_time_slot(
        scheduler.slug, timeslot_to_book
    )
    request = mocked_responses.calls[0].request
    assert (
        request.url == "https://api.schedule.nylas.com/schedule/py_example_1/timeslots"
    )
    assert request.method == responses.POST
    assert json.loads(request.body) == {
        "additional_emails": [],
        "additional_values": {
            "test": "yes",
        },
        "email": "recipient@example.com",
        "locale": "en_US",
        "name": "Recipient Doe",
        "timezone": "America/New_York",
        "slot": {
            "account_id": "test-account-id",
            "calendar_id": "test-calendar-id",
            "emails": ["recipient@example.com"],
            "host_name": "www.nylas.com",
            "start": 1636728347,
            "end": 1636731958,
        },
    }
    assert booking_response.account_id == "test-account-id"
    assert booking_response.calendar_id == "test-calendar-id"
    assert booking_response.additional_field_values == {
        "test": "yes",
    }
    assert booking_response.calendar_event_id == "test-event-id"
    assert booking_response.calendar_id == "test-calendar-id"
    assert booking_response.calendar_event_id == "test-event-id"
    assert booking_response.edit_hash == "test-edit-hash"
    assert booking_response.id == 123
    assert booking_response.is_confirmed is False
    assert booking_response.location == "Earth"
    assert booking_response.title == "Test Booking"
    assert booking_response.recipient_email == "recipient@example.com"
    assert booking_response.recipient_locale == "en_US"
    assert booking_response.recipient_name == "Recipient Doe"
    assert booking_response.recipient_tz == "America/New_York"
    assert booking_response.end_time == datetime.utcfromtimestamp(1636731958)
    assert booking_response.start_time == datetime.utcfromtimestamp(1636728347)


@pytest.mark.usefixtures("mock_scheduler_timeslots")
def test_scheduler_confirm_booking(mocked_responses, api_client):
    scheduler = blank_scheduler_page(api_client)
    booking_confirmation = api_client.scheduler.confirm_booking(
        scheduler.slug, "test-edit-hash"
    )
    request = mocked_responses.calls[0].request
    assert (
        request.url
        == "https://api.schedule.nylas.com/schedule/py_example_1/test-edit-hash/confirm"
    )
    assert request.method == responses.POST
    assert booking_confirmation.account_id == "test-account-id"
    assert booking_confirmation.calendar_id == "test-calendar-id"
    assert booking_confirmation.additional_field_values == {
        "test": "yes",
    }
    assert booking_confirmation.calendar_event_id == "test-event-id"
    assert booking_confirmation.calendar_id == "test-calendar-id"
    assert booking_confirmation.calendar_event_id == "test-event-id"
    assert booking_confirmation.edit_hash == "test-edit-hash"
    assert booking_confirmation.id == 123
    assert booking_confirmation.is_confirmed is True
    assert booking_confirmation.location == "Earth"
    assert booking_confirmation.title == "Test Booking"
    assert booking_confirmation.recipient_email == "recipient@example.com"
    assert booking_confirmation.recipient_locale == "en_US"
    assert booking_confirmation.recipient_name == "Recipient Doe"
    assert booking_confirmation.recipient_tz == "America/New_York"
    assert booking_confirmation.end_time == datetime.utcfromtimestamp(1636731958)
    assert booking_confirmation.start_time == datetime.utcfromtimestamp(1636728347)


@pytest.mark.usefixtures("mock_scheduler_timeslots")
def test_scheduler_cancel_booking(mocked_responses, api_client):
    scheduler = blank_scheduler_page(api_client)
    timeslots = api_client.scheduler.cancel_booking(
        scheduler.slug, "test-edit-hash", "It was a test."
    )
    request = mocked_responses.calls[0].request
    assert (
        request.url
        == "https://api.schedule.nylas.com/schedule/py_example_1/test-edit-hash/cancel"
    )
    assert request.method == responses.POST
    assert json.loads(request.body) == {"reason": "It was a test."}



import pytest
from urlobject import URLObject

from nylas.client.delta_models import Deltas, Delta
from nylas.client.restful_models import (
    Contact,
    File,
    Message,
    Draft,
    Thread,
    Event,
    Folder,
    Label,
)


@pytest.mark.usefixtures("mock_deltas_since")
def test_deltas_since(mocked_responses, api_client):
    deltas = api_client.deltas.since("cursor")
    request = mocked_responses.calls[0].request
    assert URLObject(request.url).path == "/delta"
    assert URLObject(request.url).query_dict == {"cursor": "cursor"}
    assert request.method == "GET"
    assert isinstance(deltas, Deltas)
    assert deltas.cursor_start == "start_cursor"
    assert deltas.cursor_end == "end_cursor"
    assert len(deltas.deltas) == 8
    assert isinstance(deltas.deltas[0].attributes, Contact)
    assert deltas.deltas[0].cursor == "contact_cursor"
    assert deltas.deltas[0].event == "create"
    assert deltas.deltas[0].id == "delta-1"
    assert deltas.deltas[0].object == "contact"
    assert isinstance(deltas.deltas[1].attributes, File)
    assert deltas.deltas[1].cursor == "file_cursor"
    assert deltas.deltas[1].event == "create"
    assert deltas.deltas[1].id == "delta-2"
    assert deltas.deltas[1].object == "file"
    assert isinstance(deltas.deltas[2].attributes, Message)
    assert deltas.deltas[2].cursor == "message_cursor"
    assert deltas.deltas[2].event == "create"
    assert deltas.deltas[2].id == "delta-3"
    assert deltas.deltas[2].object == "message"
    assert isinstance(deltas.deltas[3].attributes, Draft)
    assert deltas.deltas[3].cursor == "draft_cursor"
    assert deltas.deltas[3].event == "create"
    assert deltas.deltas[3].id == "delta-4"
    assert deltas.deltas[3].object == "draft"
    assert isinstance(deltas.deltas[4].attributes, Thread)
    assert deltas.deltas[4].cursor == "thread_cursor"
    assert deltas.deltas[4].event == "create"
    assert deltas.deltas[4].id == "delta-5"
    assert deltas.deltas[4].object == "thread"
    assert isinstance(deltas.deltas[5].attributes, Event)
    assert deltas.deltas[5].cursor == "event_cursor"
    assert deltas.deltas[5].event == "create"
    assert deltas.deltas[5].id == "delta-6"
    assert deltas.deltas[5].object == "event"
    assert isinstance(deltas.deltas[6].attributes, Folder)
    assert deltas.deltas[6].cursor == "folder_cursor"
    assert deltas.deltas[6].event == "create"
    assert deltas.deltas[6].id == "delta-7"
    assert deltas.deltas[6].object == "folder"
    assert isinstance(deltas.deltas[7].attributes, Label)
    assert deltas.deltas[7].cursor == "label_cursor"
    assert deltas.deltas[7].event == "create"
    assert deltas.deltas[7].id == "delta-8"
    assert deltas.deltas[7].object == "label"


@pytest.mark.usefixtures("mock_delta_cursor")
def test_delta_cursor(mocked_responses, api_client):
    cursor = api_client.deltas.latest_cursor()
    request = mocked_responses.calls[0].request
    assert URLObject(request.url).path == "/delta/latest_cursor"
    assert request.method == "POST"
    assert cursor == "cursor"


def evaluate_contact_delta(delta):
    assert isinstance(delta, Delta)
    assert isinstance(delta.attributes, Contact)
    assert delta.cursor == "contact_cursor"
    assert delta.event == "create"
    assert delta.id == "delta-1"
    assert delta.object == "contact"


@pytest.mark.usefixtures("mock_delta_stream")
def test_delta_streaming(mocked_responses, api_client):
    streaming = api_client.deltas.stream("cursor")
    request = mocked_responses.calls[0].request
    assert URLObject(request.url).path == "/delta/streaming"
    assert URLObject(request.url).query_dict == {"cursor": "cursor"}
    assert request.method == "GET"
    assert len(streaming) == 1
    evaluate_contact_delta(streaming[0])


@pytest.mark.usefixtures("mock_delta_stream")
def test_delta_longpoll(mocked_responses, api_client):
    longpoll = api_client.deltas.longpoll("cursor", 30)
    request = mocked_responses.calls[0].request
    assert URLObject(request.url).path == "/delta/longpoll"
    assert URLObject(request.url).query_dict == {"cursor": "cursor", "timeout": "30"}
    assert request.method == "GET"
    assert isinstance(longpoll, Deltas)
    assert longpoll.cursor_start == "start_cursor"
    assert longpoll.cursor_end == "end_cursor"
    assert len(longpoll.deltas) == 1
    evaluate_contact_delta(longpoll.deltas[0])


@pytest.mark.usefixtures("mock_delta_stream")
def test_delta_callback(mocked_responses, api_client):
    api_client.deltas.stream("cursor", callback=evaluate_contact_delta)


@pytest.mark.usefixtures("mock_delta_stream")
def test_delta_optional_params(mocked_responses, api_client):
    api_client.deltas.longpoll(
        "cursor", 30, view="expanded", include_types=["event", "file"]
    )
    request = mocked_responses.calls[0].request
    assert URLObject(request.url).path == "/delta/longpoll"
    assert URLObject(request.url).query_dict == {
        "cursor": "cursor",
        "timeout": "30",
        "view": "expanded",
        "include_types": "event,file",
    }
    assert request.method == "GET"


@pytest.mark.usefixtures("mock_delta_stream")
def test_delta_type_string(mocked_responses, api_client):
    api_client.deltas.longpoll("cursor", 30, view="expanded", excluded_types="event")
    request = mocked_responses.calls[0].request
    assert URLObject(request.url).path == "/delta/longpoll"
    assert URLObject(request.url).query_dict == {
        "cursor": "cursor",
        "timeout": "30",
        "view": "expanded",
        "excluded_types": "event",
    }
    assert request.method == "GET"


@pytest.mark.usefixtures("mock_delta_stream")
def test_delta_set_both_types_raise_error(api_client):
    with pytest.raises(ValueError) as excinfo:
        api_client.deltas.longpoll(
            "cursor",
            30,
            view="expanded",
            excluded_types="event",
            include_types="file",
        )
    assert "You cannot set both include_types and excluded_types" in str(excinfo)



from nylas.client.restful_models import (
    RestfulModel,
    NylasAPIObject,
    Contact,
    File,
    Message,
    Draft,
    Thread,
    Event,
    Folder,
    Label,
)


class Deltas(RestfulModel):
    attrs = (
        "cursor_start",
        "cursor_end",
        "_deltas",
    )
    read_only_attrs = tuple(attrs)

    def __init__(self, api):
        RestfulModel.__init__(self, Deltas, api)

    @property
    def deltas(self):
        """
        Instantiate a Delta object from the API response

        Returns:
            list[Delta]: List of Delta instantiated objects
        """
        if self._deltas:
            deltas = []
            for delta in self._deltas:
                deltas.append(Delta.create(self.api, **delta))
            return deltas


class Delta(RestfulModel):
    attrs = (
        "id",
        "cursor",
        "event",
        "object",
        "_attributes",
    )
    read_only_attrs = tuple(attrs)
    class_mapping = {
        "contact": Contact,
        "file": File,
        "message": Message,
        "draft": Draft,
        "thread": Thread,
        "event": Event,
        "folder": Folder,
        "label": Label,
    }

    def __init__(self, api):
        RestfulModel.__init__(self, Delta, api)

    @property
    def attributes(self):
        """
        Instantiate the object provided in the Delta

        Returns:
            NylasAPIObject: The object of NylasAPIObject type represented in the Delta
        """
        if self._attributes and self.object and self.object in self.class_mapping:
            return self.class_mapping[self.object].create(self.api, **self._attributes)



from __future__ import division
from datetime import datetime, timedelta
from enum import Enum


def timestamp_from_dt(dt, epoch=datetime(1970, 1, 1)):
    """
    Convert a datetime to a timestamp.
    https://stackoverflow.com/a/8778548/141395
    """
    # For offset-aware datetime objects, convert them first before performing delta
    if dt.tzinfo is not None and dt.utcoffset() is not None:
        dt = dt.replace(tzinfo=None) - dt.utcoffset()

    delta = dt - epoch

    return int(delta.total_seconds() / timedelta(seconds=1).total_seconds())


def create_request_body(data, datetime_attrs):
    """
    Given a dictionary of data, and a dictionary of datetime attributes,
    return a new dictionary that is suitable for a request. It converts
    any datetime attributes that may be present to their timestamped
    equivalent, and it filters out any attributes set to "None".
    """
    if not data:
        return data

    new_data = {}
    for key, value in data.items():
        if key in datetime_attrs and isinstance(value, datetime):
            new_key = datetime_attrs[key]
            new_data[new_key] = timestamp_from_dt(value)
        elif value is not None:
            new_data[key] = value

    return new_data


def convert_metadata_pairs_to_array(data):
    """
    Given a dictionary of metadata pairs, convert it to key-value pairs
    in the format the Nylas API expects: "events?metadata_pair=<key>:<value>"
    """
    if not data:
        return data

    metadata_pair = []
    for key, value in data.items():
        metadata_pair.append(key + ":" + value)

    return metadata_pair


class AuthMethod(str, Enum):
    """
    This is an Enum representing all the different authentication methods that the Nylas APIs support
    """

    BEARER = 1
    BASIC = 2
    BASIC_CLIENT_ID_AND_SECRET = 3


class HttpMethod(str, Enum):
    """
    This is an Enum representing all the HTTP Methods that the Nylas APIs support
    """

    GET = 1
    PUT = 2
    POST = 3
    PATCH = 4
    DELETE = 5



from datetime import datetime

from nylas.client.restful_models import RestfulModel, Draft
from nylas.utils import timestamp_from_dt


class OutboxMessage(RestfulModel):
    attrs = Draft.attrs + [
        "send_at",
        "retry_limit_datetime",
        "original_send_at",
    ]
    datetime_attrs = {
        "send_at": "send_at",
        "retry_limit_datetime": "retry_limit_datetime",
        "original_send_at": "original_send_at",
    }
    read_only_attrs = {"send_at", "retry_limit_datetime", "original_send_at"}
    collection_name = "v2/outbox"

    def __init__(self, api):
        RestfulModel.__init__(self, OutboxMessage, api)


class OutboxJobStatus(RestfulModel):
    attrs = [
        "account_id",
        "job_status_id",
        "status",
        "original_data",
    ]
    collection_name = "v2/outbox"

    def __init__(self, api):
        RestfulModel.__init__(self, OutboxJobStatus, api)


class SendGridVerifiedStatus(RestfulModel):
    attrs = [
        "domain_verified",
        "sender_verified",
    ]
    collection_name = "v2/outbox/onboard"

    def __init__(self, api):
        RestfulModel.__init__(self, SendGridVerifiedStatus, api)


class Outbox:
    def __init__(self, api):
        self.api = api

    def send(self, draft, send_at, retry_limit_datetime=None):
        """
        Send a message via Outbox

        Args:
            draft (Draft | OutboxMessage): The message to send
            send_at (datetime | int): The date and time to send the message. If set to 0, Outbox will send this message immediately.
            retry_limit_datetime (datetime | int): Optional date and time to stop retry attempts for a message.

        Returns:
            OutboxJobStatus: The Outbox message job status

        Raises:
            ValueError: If the date and times provided are not valid
        """
        draft_json = draft.as_json()
        send_at, retry_limit_datetime = self._validate_and_format_datetime(
            send_at, retry_limit_datetime
        )

        draft_json["send_at"] = send_at
        if retry_limit_datetime is not None:
            draft_json["retry_limit_datetime"] = retry_limit_datetime

        return self.api._create_resource(OutboxJobStatus, draft_json)

    def update(
        self, job_status_id, draft=None, send_at=None, retry_limit_datetime=None
    ):
        """
        Update a scheduled Outbox message

        Args:
            job_status_id (str): The ID of the outbox job status
            draft (Draft | OutboxMessage): The message object with updated values
            send_at (datetime | int): The date and time to send the message. If set to 0, Outbox will send this message immediately.
            retry_limit_datetime (datetime | int): Optional date and time to stop retry attempts for a message.

        Returns:
            OutboxJobStatus: The updated Outbox message job status

        Raises:
            ValueError: If the date and times provided are not valid
        """
        payload = {}
        if draft:
            payload = draft.as_json()
        send_at, retry_limit_datetime = self._validate_and_format_datetime(
            send_at, retry_limit_datetime
        )

        if send_at is not None:
            payload["send_at"] = send_at
        if retry_limit_datetime is not None:
            payload["retry_limit_datetime"] = retry_limit_datetime

        response = self.api._patch_resource(OutboxJobStatus, job_status_id, payload)
        return OutboxJobStatus.create(self.api, **response)

    def delete(self, job_status_id):
        """
        Delete a scheduled Outbox message

        Args:
            job_status_id (str): The ID of the outbox job status to delete
        """

        self.api._delete_resource(OutboxJobStatus, job_status_id)

    def send_grid_verification_status(self):
        """
        SendGrid - Check Authentication and Verification Status

        Returns:
            SendGridVerifiedStatus: The status of the domain authentication and the single sender verification for SendGrid integrations

        Raises:
            RuntimeError: If the server returns an object without results
        """
        response = self.api._get_resource_raw(
            SendGridVerifiedStatus, None, extra="verified_status"
        )
        response_body = response.json()
        if "results" not in response_body:
            raise RuntimeError(
                "Unexpected response from the API server. Returned 200 but no 'ics' string found."
            )
        return SendGridVerifiedStatus.create(self.api, **response_body["results"])

    def delete_send_grid_sub_user(self, email_address):
        """
        SendGrid -  Delete SendGrid Subuser and UAS Grant

        Args:
            email_address (str): Email address for SendGrid subuser to delete
        """
        payload = {"email": email_address}

        self.api._delete_resource(SendGridVerifiedStatus, "subuser", data=payload)

    def _validate_and_format_datetime(self, send_at, retry_limit_datetime):
        send_at_epoch = (
            timestamp_from_dt(send_at) if isinstance(send_at, datetime) else send_at
        )
        retry_limit_datetime_epoch = (
            timestamp_from_dt(retry_limit_datetime)
            if isinstance(retry_limit_datetime, datetime)
            else retry_limit_datetime
        )
        now_epoch = timestamp_from_dt(datetime.today())

        if send_at_epoch and send_at_epoch != 0 and send_at_epoch < now_epoch:
            raise ValueError(
                "Cannot set message to be sent at a time before the current time."
            )

        if retry_limit_datetime_epoch and retry_limit_datetime_epoch != 0:
            current_send_at = (
                send_at_epoch if send_at_epoch and send_at_epoch != 0 else now_epoch
            )
            if retry_limit_datetime_epoch < current_send_at:
                raise ValueError(
                    "Cannot set message to stop retrying before time to send at."
                )

        return send_at_epoch, retry_limit_datetime_epoch



from datetime import datetime, timedelta
import json

import pytest
from urlobject import URLObject

from nylas.utils import timestamp_from_dt


@pytest.mark.usefixtures("mock_outbox")
def test_outbox_send(mocked_responses, api_client):
    draft, tomorrow, day_after = prepare_outbox_request(api_client)

    job_status = api_client.outbox.send(draft, tomorrow, retry_limit_datetime=day_after)

    request = mocked_responses.calls[0].request
    assert URLObject(request.url).path == "/v2/outbox"
    assert request.method == "POST"
    body = json.loads(request.body)
    evaluate_message(body, tomorrow, day_after)
    assert job_status["job_status_id"] == "job-status-id"
    assert job_status["status"] == "pending"
    assert job_status["account_id"] == "account-id"
    original_data = job_status["original_data"]
    evaluate_message(original_data, tomorrow, day_after)
    assert original_data["original_send_at"] == timestamp_from_dt(tomorrow)


@pytest.mark.usefixtures("mock_outbox")
def test_outbox_update(mocked_responses, api_client):
    draft, tomorrow, day_after = prepare_outbox_request(api_client)

    api_client.outbox.update(
        "job-status-id", draft=draft, send_at=tomorrow, retry_limit_datetime=day_after
    )

    request = mocked_responses.calls[0].request
    assert URLObject(request.url).path == "/v2/outbox/job-status-id"
    assert request.method == "PATCH"
    body = json.loads(request.body)
    evaluate_message(body, tomorrow, day_after)


@pytest.mark.usefixtures("mock_outbox")
def test_outbox_send_at_before_today_should_raise(mocked_responses, api_client):
    with pytest.raises(ValueError) as excinfo:
        api_client.outbox._validate_and_format_datetime(636309514, None)
    assert "Cannot set message to be sent at a time before the current time." in str(
        excinfo
    )


@pytest.mark.usefixtures("mock_outbox")
def test_outbox_retry_limit_datetime_before_send_at_should_raise(
    mocked_responses, api_client
):
    tomorrow = datetime.today() + timedelta(days=1)
    day_after = tomorrow + timedelta(days=1)
    with pytest.raises(ValueError) as excinfo:
        api_client.outbox._validate_and_format_datetime(
            send_at=day_after, retry_limit_datetime=tomorrow
        )
    assert "Cannot set message to stop retrying before time to send at." in str(excinfo)


@pytest.mark.usefixtures("mock_outbox")
def test_outbox_retry_limit_datetime_before_today_should_raise(
    mocked_responses, api_client
):
    with pytest.raises(ValueError) as excinfo:
        api_client.outbox._validate_and_format_datetime(None, 636309514)
    assert "Cannot set message to stop retrying before time to send at." in str(excinfo)


@pytest.mark.usefixtures("mock_outbox")
def test_outbox_delete(mocked_responses, api_client):
    api_client.outbox.delete("job-status-id")

    request = mocked_responses.calls[0].request
    assert URLObject(request.url).path == "/v2/outbox/job-status-id"
    assert request.method == "DELETE"


@pytest.mark.usefixtures("mock_outbox_send_grid")
def test_outbox_send_grid_verification(mocked_responses, api_client):
    verification_status = api_client.outbox.send_grid_verification_status()

    request = mocked_responses.calls[0].request
    assert URLObject(request.url).path == "/v2/outbox/onboard/verified_status"
    assert request.method == "GET"
    assert verification_status.domain_verified is True
    assert verification_status.sender_verified is True


@pytest.mark.usefixtures("mock_outbox_send_grid")
def test_outbox_send_grid_verification(mocked_responses, api_client):
    api_client.outbox.delete_send_grid_sub_user("test@email.com")

    request = mocked_responses.calls[0].request
    assert URLObject(request.url).path == "/v2/outbox/onboard/subuser"
    assert request.method == "DELETE"


# Test helpers


def prepare_outbox_request(api_client):
    draft = api_client.drafts.create()
    draft.subject = "With Love, from Nylas"
    draft.to = [{"email": "test@email.com", "name": "Me"}]
    draft.body = "This email was sent using the Nylas email API. Visit https://nylas.com for details."
    tomorrow = datetime.today() + timedelta(days=1)
    day_after = tomorrow + timedelta(days=1)

    return draft, tomorrow, day_after


def evaluate_message(message, send_at, retry_limit_datetime):
    assert message["to"] == [{"email": "test@email.com", "name": "Me"}]
    assert message["subject"] == "With Love, from Nylas"
    assert (
        message["body"]
        == "This email was sent using the Nylas email API. Visit https://nylas.com for details."
    )
    assert message["send_at"] == timestamp_from_dt(send_at)
    assert message["retry_limit_datetime"] == timestamp_from_dt(retry_limit_datetime)



import json
from datetime import datetime

import pytest
from requests import RequestException
from nylas.utils import timestamp_from_dt

# pylint: disable=len-as-condition


@pytest.mark.usefixtures("mock_drafts")
def test_draft_attrs(api_client):
    draft = api_client.drafts.first()
    expected_modified = datetime(2015, 8, 4, 10, 34, 46)
    assert draft.last_modified_at == expected_modified
    assert draft.date == timestamp_from_dt(expected_modified)


@pytest.mark.usefixtures("mock_draft_saved_response", "mock_draft_sent_response")
def test_save_send_draft(api_client):
    draft = api_client.drafts.create()
    draft.to = [{"name": "My Friend", "email": "my.friend@example.com"}]
    draft.subject = "Here's an attachment"
    draft.body = "Cheers mate!"
    draft.save()

    draft.subject = "Stay polish, stay hungary"
    draft.save()
    assert draft.subject == "Stay polish, stay hungary"

    msg = draft.send()
    assert msg["thread_id"] == "clm33kapdxkposgltof845v9s"

    # Second time should throw an error
    with pytest.raises(RequestException):
        draft.send()


def test_draft_as_json(api_client):
    draft = api_client.drafts.create()
    draft.to = [{"name": "My Friend", "email": "my.friend@example.com"}]
    draft.subject = "Here's an attachment"
    draft.body = "Cheers mate!"
    draft.from_ = [{"name": "Me", "email": "me@example.com"}]
    msg = draft.as_json()
    assert msg["to"] == [{"name": "My Friend", "email": "my.friend@example.com"}]
    assert msg["subject"] == "Here's an attachment"
    assert msg["body"] == "Cheers mate!"
    assert msg["from"] == [{"name": "Me", "email": "me@example.com"}]


@pytest.mark.usefixtures("mock_draft_saved_response", "mock_draft_sent_response")
def test_save_send_draft_with_tracking(mocked_responses, api_client):
    tracking = {
        "links": "true",
        "opens": "true",
        "thread_replies": "true",
        "payload": "new-payload",
    }
    draft = api_client.drafts.create()
    draft.to = [{"name": "My Friend", "email": "my.friend@example.com"}]
    draft.subject = "Here's an attachment"
    draft.body = "Cheers mate!"
    draft.save()

    draft.tracking = tracking
    draft.save()
    assert draft.tracking == tracking

    draft.send()
    send_payload = json.loads(mocked_responses.calls[-1].request.body)
    assert send_payload["tracking"] == tracking


@pytest.mark.usefixtures("mock_draft_send_unsaved_response")
def test_send_draft_with_tracking(mocked_responses, api_client):
    tracking = {"opens": "true", "payload": "payload"}
    draft = api_client.drafts.create()
    draft.to = [{"name": "My Friend", "email": "my.friend@example.com"}]
    draft.subject = "Newsletter"
    draft.body = "Our latest sale!"
    draft.tracking = tracking
    draft.send()

    send_payload = json.loads(mocked_responses.calls[-1].request.body)
    assert send_payload["tracking"] == tracking


@pytest.mark.usefixtures("mock_files")
def test_draft_attachment(api_client):
    draft = api_client.drafts.create()
    attachment = api_client.files.create()
    attachment.filename = "dummy"
    attachment.data = "data"

    assert len(draft.file_ids) == 0
    draft.attach(attachment)
    assert len(draft.file_ids) == 1
    assert attachment.id in draft.file_ids

    unattached = api_client.files.create()
    unattached.filename = "unattached"
    unattached.data = "foo"
    draft.detach(unattached)
    assert len(draft.file_ids) == 1
    assert attachment.id in draft.file_ids
    assert unattached.id not in draft.file_ids

    draft.detach(attachment)
    assert len(draft.file_ids) == 0


@pytest.mark.usefixtures("mock_draft_saved_response", "mock_draft_deleted_response")
def test_delete_draft(api_client):
    draft = api_client.drafts.create()
    # Unsaved draft shouldn't throw an error on .delete(), but won't actually
    # delete anything.
    draft.delete()
    # Now save the draft...
    draft.save()
    # ... and delete it for real.
    draft.delete()


@pytest.mark.usefixtures("mock_draft_saved_response")
def test_draft_version(api_client):
    draft = api_client.drafts.create()
    assert "version" not in draft
    draft.save()
    assert draft["version"] == 0
    draft.update()
    assert draft["version"] == 1
    draft.update()
    assert draft["version"] == 2



from copy import copy

from nylas.client.restful_model_collection import RestfulModelCollection, CHUNK_SIZE
from nylas.client.restful_models import NylasAPIObject
from nylas.utils import AuthMethod
from enum import Enum


class Integration(NylasAPIObject):
    attrs = (
        "name",
        "provider",
        "expires_in",
        "settings",
        "redirect_uris",
        "scope",
        "id",
    )
    read_only_attrs = {"provider", "id"}
    auth_method = AuthMethod.BASIC_CLIENT_ID_AND_SECRET
    collection_name = "connect/integrations"

    def __init__(self, api):
        NylasAPIObject.__init__(self, Integration, api)
        self.settings = {}
        self.scope = []

    def set_client_id(self, client_id):
        """
        Set the client ID of the OAuth provider

        Args:
            client_id (str): Client ID of the OAuth provider
        """
        self.settings["client_id"] = client_id

    def set_client_secret(self, client_secret):
        """
        Set the client secret of the OAuth provider

        Args:
            client_secret (str): Client secret of the OAuth provider
        """
        self.settings["client_secret"] = client_secret

    @classmethod
    def create(cls, api, **kwargs):
        if "data" in kwargs:
            kwargs = kwargs.get("data")
        obj = super(Integration, cls).create(api, **kwargs)
        if "provider" in kwargs:
            obj["id"] = kwargs.get("provider")

        return obj

    def as_json(self, enforce_read_only=True):
        dct = super(Integration, self).as_json(enforce_read_only)
        if enforce_read_only is False:
            return dct

        if not self.id:
            if isinstance(self.provider, Authentication.Provider):
                dct["provider"] = self.provider.value
            else:
                dct["provider"] = self.provider

        return dct

    def _update_resource(self, **kwargs):
        provider = self.id or self.provider
        return self.api._patch_resource(self.cls, provider, self.as_json(), **kwargs)


class Grant(NylasAPIObject):
    attrs = (
        "id",
        "provider",
        "state",
        "email",
        "ip",
        "grant_status",
        "user_agent",
        "created_at",
        "updated_at",
        "settings",
        "metadata",
        "scope",
    )
    read_only_attrs = {
        "id",
        "email",
        "ip",
        "grant_status",
        "user_agent",
        "created_at",
        "updated_at",
    }
    auth_method = AuthMethod.BASIC_CLIENT_ID_AND_SECRET
    collection_name = "connect/grants"

    def __init__(self, api):
        NylasAPIObject.__init__(self, Grant, api)
        self.settings = {}
        self.metadata = {}
        self.scope = []

    @classmethod
    def create(cls, api, **kwargs):
        if "data" in kwargs:
            kwargs = kwargs.get("data")
        obj = super(Grant, cls).create(api, **kwargs)
        return obj

    def as_json(self, enforce_read_only=True):
        dct = super(Grant, self).as_json(enforce_read_only)
        if enforce_read_only is False:
            return dct

        # provider and state can not be updated
        if self.id:
            del dct["provider"]
            del dct["state"]
        else:
            if isinstance(self.provider, Authentication.Provider):
                dct["provider"] = self.provider.value
            else:
                dct["provider"] = self.provider

        return dct

    def _update_resource(self, **kwargs):
        return self.api._patch_resource(self.cls, self.id, self.as_json(), **kwargs)


class Authentication(object):
    def __init__(self, api):
        self._app_name = "beta"
        self._region = Authentication.Region.US
        # Make a copy of the API as we need to change the base url for Integration calls
        self.api = copy(api)
        self._set_integrations_api_url()

    @property
    def app_name(self):
        return self._app_name

    @app_name.setter
    def app_name(self, value):
        """
        Set the name of the application to prefix the URL for all integration calls for this instance

        Args:
            value (str): The name of the application
        """
        self._app_name = value
        self._set_integrations_api_url()

    @property
    def region(self):
        return self._region

    @region.setter
    def region(self, value):
        """
        Set the region to prefix the URL for all integration calls for this instance

        Args:
            value (Integration.Region): The region
        """
        self._region = value
        self._set_integrations_api_url()

    @property
    def integrations(self):
        """
        Integrations API for integrating a provider to the Nylas application

        Returns:
            IntegrationRestfulModelCollection: The Integration API configured with the app_name and region
        """
        return IntegrationRestfulModelCollection(self.api)

    @property
    def grants(self):
        """
        Native Authentication for the integrated provider

        Returns:
            GrantRestfulModelCollection: The Grants API configured with the app_name and region
        """
        return GrantRestfulModelCollection(self.api)

    def hosted_authentication(
        self,
        provider,
        redirect_uri,
        grant_id=None,
        login_hint=None,
        state=None,
        expires_in=None,
        settings=None,
        metadata=None,
        scope=None,
    ):
        """
        Hosted Authentication for the integrated provider

        Args:
            provider (Authentication.Provider): OAuth provider
            redirect_uri (str): The URI for the final redirect
            grant_id (str): Existing Grant ID to trigger a re-authentication
            login_hint (str): Hint to simplify the login flow
            state (str): State value to return after authentication flow is completed
            expires_in (int): How long this request (and the attached login) ID will remain valid before the link expires
            settings (dict[str, str]): Settings required by provider
            metadata (dict[str, any]): Metadata to store as part of the grant
            scope (list[str]): OAuth provider-specific scopes

        Returns:
            dict[str, any]: The login information
        """
        request = {"provider": provider, "redirect_uri": redirect_uri}
        if grant_id:
            request["grant_id"] = grant_id
        if login_hint:
            request["login_hint"] = login_hint
        if state:
            request["state"] = state
        if expires_in:
            request["expires_in"] = expires_in
        if settings:
            request["settings"] = settings
        if metadata:
            request["metadata"] = metadata
        if scope:
            request["scope"] = scope

        response = self.api._post_resource(Grant, "auth", None, request, path="connect")
        if "data" in response:
            response = response["data"]

        return response

    def _set_integrations_api_url(self):
        self.api.api_server = "https://{app_name}.{region}.nylas.com".format(
            app_name=self.app_name, region=self.region.value
        )

    def _hosted_authentication_enhanced_events(
        self, provider, redirect_uri, account_id
    ):
        request = {
            "provider": provider,
            "redirect_uri": redirect_uri,
            "account_id": account_id,
        }
        response = self.api._post_resource(Grant, "auth", None, request, path="connect")
        if "data" in response:
            response = response["data"]

        return response

    class Region(str, Enum):
        """
        This is an Enum the regions supported by the Integrations API
        """

        US = "us"
        EU = "eu"

    class Provider(str, Enum):
        """
        This is an Enum representing all the available providers for integrations
        """

        GOOGLE = "google"
        MICROSOFT = "microsoft"
        IMAP = "imap"
        ZOOM = "zoom"


class AuthenticationRestfulModelCollection(RestfulModelCollection):
    def __init__(self, model_class, api):
        RestfulModelCollection.__init__(self, model_class, api)

    def _get_model_collection(self, offset=0, limit=CHUNK_SIZE):
        filters = copy(self.filters)
        filters["offset"] = offset
        if not filters.get("limit"):
            filters["limit"] = limit

        response = self.api._get_resource_raw(self.model_class, None, **filters).json()
        if "data" not in response or response["data"] is None:
            return []

        return [
            self.model_class.create(self, **x)
            for x in response["data"]
            if x is not None
        ]


class IntegrationRestfulModelCollection(AuthenticationRestfulModelCollection):
    def __init__(self, api):
        AuthenticationRestfulModelCollection.__init__(self, Integration, api)

    def get(self, provider):
        """
        Get an existing integration for a provider

        Args:
            provider (Authentication.Provider): The provider

        Returns:
            Integration: The existing integration
        """
        return super(IntegrationRestfulModelCollection, self).get(provider.value)

    def delete(self, provider, data=None, **kwargs):
        """
        Deletes an existing integration for a provider

        Args:
            provider (Authentication.Provider): The provider
        """
        super(IntegrationRestfulModelCollection, self).delete(
            provider.value, data=data, **kwargs
        )


class GrantRestfulModelCollection(AuthenticationRestfulModelCollection):
    def __init__(self, api):
        AuthenticationRestfulModelCollection.__init__(self, Grant, api)

    def on_demand_sync(self, grant_id, sync_from=None):
        """
        Trigger a grant sync on demand

        Args:
            grant_id (str): The grant ID to sync
            sync_from (int): Epoch timestamp when the sync starts from

        Returns:
            Grant: The grant after triggering the sync
        """
        path = "sync"
        if sync_from:
            path = path + "?sync_from={}".format(sync_from)
        response = self.api._post_resource(Grant, grant_id, path, data=None)
        return self.model_class.create(self, **response)



from datetime import datetime

import pytest
from nylas.client.restful_models import JobStatus


@pytest.mark.usefixtures("mock_job_statuses")
def test_first_job_status(api_client):
    job_status = api_client.job_statuses.first()
    assert isinstance(job_status, JobStatus)


@pytest.mark.usefixtures("mock_job_statuses")
def test_all_job_status(api_client):
    job_statuses = api_client.job_statuses.all()
    assert len(job_statuses) == 2
    for job_status in job_statuses:
        assert isinstance(job_status, JobStatus)


@pytest.mark.usefixtures("mock_job_statuses")
def test_job_status(api_client):
    job_status = api_client.job_statuses.first()
    assert job_status["account_id"] == "test_account_id"
    assert job_status["action"] == "save_draft"
    assert job_status["id"] == "test_id"
    assert job_status["job_status_id"] == "test_job_status_id"
    assert job_status["object"] == "message"
    assert job_status["status"] == "successful"
    assert job_status["created_at"] == datetime(2021, 6, 4, 22, 36)
    assert job_status["metadata"] == {"message_id": "nylas_message_id"}


@pytest.mark.usefixtures("mock_job_statuses")
def test_job_status_is_successful(api_client):
    job_status = api_client.job_statuses.first()
    assert job_status.is_successful() is True


@pytest.mark.usefixtures("mock_job_statuses")
def test_job_status_is_successful_false(api_client):
    job_status = api_client.job_statuses.first()
    job_status.status = "failed"
    assert job_status.is_successful() is False



import pytest
from nylas.client.restful_models import RoomResource


@pytest.mark.usefixtures("mock_resources")
def test_first_resource(api_client):
    resource = api_client.room_resources.first()
    assert isinstance(resource, RoomResource)


@pytest.mark.usefixtures("mock_resources")
def test_all_resources(api_client):
    resources = api_client.room_resources.all()
    assert len(resources) == 2
    for resource in resources:
        assert isinstance(resource, RoomResource)


@pytest.mark.usefixtures("mock_resources")
def test_resource(api_client):
    resource = api_client.room_resources.first()
    assert resource["object"] == "room_resource"
    assert resource["email"] == "training-room-1A@google.com"
    assert resource["name"] == "Google Training Room"
    assert resource["building"] == "San Francisco"
    assert resource["capacity"] == "10"
    assert resource["floor_name"] == "7"
    assert resource["floor_number"] is None



import json
from datetime import date
import pytest
from six import binary_type
from nylas.client.restful_models import Contact


@pytest.mark.usefixtures("mock_contacts")
def test_list_contacts(api_client):
    contacts = list(api_client.contacts)
    assert len(contacts) == 3
    assert all(isinstance(x, Contact) for x in contacts)


@pytest.mark.usefixtures("mock_contact")
def test_get_contact(api_client):
    contact = api_client.contacts.get("9hga75n6mdvq4zgcmhcn7hpys")
    assert contact is not None
    assert isinstance(contact, Contact)
    assert contact.given_name == "Given"
    assert contact.surname == "Sur"
    assert contact.birthday == date(1964, 10, 5)
    assert contact.source == "inbox"


@pytest.mark.usefixtures("mock_contacts")
def test_create_contact(api_client, mocked_responses):
    contact = api_client.contacts.create()
    contact.given_name = "Monkey"
    contact.surname = "Business"
    assert not mocked_responses.calls
    contact.save()
    assert len(mocked_responses.calls) == 1
    assert contact.id is not None
    assert contact.given_name == "Monkey"
    assert contact.surname == "Business"


@pytest.mark.usefixtures("mock_contact")
def test_update_contact(api_client, mocked_responses):
    contact = api_client.contacts.get("9hga75n6mdvq4zgcmhcn7hpys")
    assert len(mocked_responses.calls) == 1
    assert contact.job_title == "QA Tester"
    contact.job_title = "Factory Owner"
    contact.office_location = "Willy Wonka Factory"
    contact.save()
    assert len(mocked_responses.calls) == 2
    assert contact.id == "9hga75n6mdvq4zgcmhcn7hpys"
    assert contact.job_title == "Factory Owner"
    assert contact.office_location == "Willy Wonka Factory"


@pytest.mark.usefixtures("mock_contact")
def test_contact_picture(api_client, mocked_responses):
    contact = api_client.contacts.get("9hga75n6mdvq4zgcmhcn7hpys")
    assert len(mocked_responses.calls) == 1
    assert contact.picture_url
    picture = contact.get_picture()
    assert len(mocked_responses.calls) == 2
    picture_call = mocked_responses.calls[1]
    assert contact.picture_url == picture_call.request.url
    assert picture.headers["Content-Type"] == "image/jpeg"
    content = picture.read()
    assert isinstance(content, binary_type)


@pytest.mark.usefixtures("mock_contacts")
def test_contact_no_picture(api_client, mocked_responses):
    contact = api_client.contacts.get("4zqkfw8k1d12h0k784ipeh498")
    assert len(mocked_responses.calls) == 1
    assert not contact.picture_url
    picture = contact.get_picture()
    assert len(mocked_responses.calls) == 1
    assert not picture


@pytest.mark.usefixtures("mock_contact")
def test_contact_emails(api_client):
    contact = api_client.contacts.get("9hga75n6mdvq4zgcmhcn7hpys")
    assert isinstance(contact.emails, dict)
    assert contact.emails["first"] == ["one@example.com"]
    assert contact.emails["second"] == ["two@example.com"]
    assert contact.emails["primary"] == ["abc@example.com", "xyz@example.com"]
    assert contact.emails[None] == ["unknown@example.com"]
    assert "absent" not in contact.emails


@pytest.mark.usefixtures("mock_contact")
def test_contact_im_addresses(api_client):
    contact = api_client.contacts.get("9hga75n6mdvq4zgcmhcn7hpys")
    assert isinstance(contact.im_addresses, dict)
    assert contact.im_addresses["aim"] == ["SmarterChild"]
    assert contact.im_addresses["gtalk"] == ["fake@gmail.com", "fake2@gmail.com"]
    assert "absent" not in contact.im_addresses


@pytest.mark.usefixtures("mock_contact")
def test_contact_physical_addresses(api_client):
    contact = api_client.contacts.get("9hga75n6mdvq4zgcmhcn7hpys")
    assert isinstance(contact.physical_addresses, dict)
    addr = contact.physical_addresses["home"][0]
    assert isinstance(addr, dict)
    assert addr["format"] == "structured"
    assert addr["street_address"] == "123 Awesome Street"
    assert "absent" not in contact.physical_addresses


@pytest.mark.usefixtures("mock_contact")
def test_contact_phone_numbers(api_client):
    contact = api_client.contacts.get("9hga75n6mdvq4zgcmhcn7hpys")
    assert isinstance(contact.phone_numbers, dict)
    assert contact.phone_numbers["home"] == ["555-555-5555"]
    assert contact.phone_numbers["mobile"] == ["555-555-5555", "987654321"]
    assert "absent" not in contact.phone_numbers


@pytest.mark.usefixtures("mock_contact")
def test_contact_web_pages(api_client):
    contact = api_client.contacts.get("9hga75n6mdvq4zgcmhcn7hpys")
    assert isinstance(contact.web_pages, dict)
    profiles = ["http://www.facebook.com/abc", "http://www.twitter.com/abc"]
    assert contact.web_pages["profile"] == profiles
    assert contact.web_pages[None] == ["http://example.com"]
    assert "absent" not in contact.web_pages


@pytest.mark.usefixtures("mock_contact")
def test_update_contact_special_values(api_client, mocked_responses):
    contact = api_client.contacts.get("9hga75n6mdvq4zgcmhcn7hpys")
    assert len(mocked_responses.calls) == 1
    contact.birthday = date(1999, 3, 6)
    contact.emails["absent"].append("absent@fake.com")
    contact.im_addresses["absent"].append("absent-im")
    contact.physical_addresses["absent"].append(
        {
            "type": "absent",
            "format": "structured",
            "street_address": "123 Absent Street",
        }
    )
    contact.phone_numbers["absent"].append("222-333-4444")
    contact.web_pages["absent"].append("http://absent.com/me")
    contact.save()
    assert len(mocked_responses.calls) == 2
    assert contact.id == "9hga75n6mdvq4zgcmhcn7hpys"
    assert contact.emails["absent"] == ["absent@fake.com"]
    assert contact.im_addresses["absent"] == ["absent-im"]
    assert contact.physical_addresses["absent"] == [
        {
            "type": "absent",
            "format": "structured",
            "street_address": "123 Absent Street",
        }
    ]
    assert contact.phone_numbers["absent"] == ["222-333-4444"]
    assert contact.web_pages["absent"] == ["http://absent.com/me"]

    request = mocked_responses.calls[-1].request
    req_body = json.loads(request.body)
    birthday = "1999-03-06"
    email_address = {"type": "absent", "email": "absent@fake.com"}
    im_address = {"type": "absent", "im_address": "absent-im"}
    physical_address = {
        "type": "absent",
        "format": "structured",
        "street_address": "123 Absent Street",
    }
    phone_number = {"type": "absent", "number": "222-333-4444"}
    web_page = {"type": "absent", "url": "http://absent.com/me"}
    assert req_body["birthday"] == birthday
    assert email_address in req_body["emails"]
    assert im_address in req_body["im_addresses"]
    assert physical_address in req_body["physical_addresses"]
    assert phone_number in req_body["phone_numbers"]
    assert web_page in req_body["web_pages"]



import copy

from nylas.client.restful_model_collection import RestfulModelCollection
from nylas.client.restful_models import Scheduler
from nylas.client.scheduler_models import (
    SchedulerTimeSlot,
    SchedulerBookingConfirmation,
)


class SchedulerRestfulModelCollection(RestfulModelCollection):
    def __init__(self, api):
        # Make a copy of the API as we need to change the base url for Scheduler calls
        scheduler_api = copy.copy(api)
        scheduler_api.api_server = "https://api.schedule.nylas.com"
        RestfulModelCollection.__init__(self, Scheduler, scheduler_api)

    def get_google_availability(self):
        return self._execute_provider_availability("google")

    def get_office_365_availability(self):
        return self._execute_provider_availability("o365")

    def get_page_slug(self, slug):
        page_response = self.api._get_resource_raw(
            self.model_class, slug, extra="info", path="schedule"
        ).json()
        return Scheduler.create(self.api, **page_response)

    def get_available_time_slots(self, slug):
        response = self.api._get_resource_raw(
            self.model_class, slug, extra="timeslots", path="schedule"
        ).json()
        return [
            SchedulerTimeSlot.create(self.api, **x) for x in response if x is not None
        ]

    def book_time_slot(self, slug, timeslot):
        response = self.api._post_resource(
            self.model_class, slug, "timeslots", timeslot.as_json(), path="schedule"
        )
        return SchedulerBookingConfirmation.create(self.api, **response)

    def cancel_booking(self, slug, edit_hash, reason):
        return self.api._post_resource(
            self.model_class,
            slug,
            "{}/cancel".format(edit_hash),
            {"reason": reason},
            path="schedule",
        )

    def confirm_booking(self, slug, edit_hash):
        booking_response = self.api._post_resource(
            self.model_class, slug, "{}/confirm".format(edit_hash), {}, path="schedule"
        )
        return SchedulerBookingConfirmation.create(self.api, **booking_response)

    def _execute_provider_availability(self, provider):
        return self.api._get_resource_raw(
            self.model_class,
            None,
            extra="availability/{}".format(provider),
            path="schedule",
        ).json()



from nylas.client.restful_models import RestfulModel


class SchedulerTimeSlot(RestfulModel):
    attrs = ["account_id", "calendar_id", "host_name", "emails"]
    datetime_attrs = {"start": "start", "end": "end"}

    def __init__(self, api):
        RestfulModel.__init__(self, SchedulerTimeSlot, api)


class SchedulerBookingConfirmation(RestfulModel):
    attrs = [
        "id",
        "account_id",
        "additional_field_values",
        "calendar_event_id",
        "calendar_id",
        "edit_hash",
        "is_confirmed",
        "location",
        "recipient_email",
        "recipient_locale",
        "recipient_name",
        "recipient_tz",
        "title",
    ]
    datetime_attrs = {"start_time": "start_time", "end_time": "end_time"}

    def __init__(self, api):
        RestfulModel.__init__(self, SchedulerBookingConfirmation, api)


class SchedulerBookingRequest(RestfulModel):
    attrs = [
        "additional_values",
        "additional_emails",
        "email",
        "locale",
        "name",
        "page_hostname",
        "replaces_booking_hash",
        "timezone",
        "slot",
    ]

    def __init__(self, api):
        RestfulModel.__init__(self, SchedulerBookingRequest, api)

    def as_json(self, enforce_read_only=True):
        dct = RestfulModel.as_json(self)
        if "additional_values" not in dct or dct["additional_values"] is None:
            dct["additional_values"] = {}
        if "additional_emails" not in dct or dct["additional_emails"] is None:
            dct["additional_emails"] = []
        if "slot" in dct and isinstance(dct["slot"], SchedulerTimeSlot):
            dct["slot"] = dct["slot"].as_json()

        return dct



import re
import json
from six.moves.urllib.parse import parse_qs  # pylint: disable=relative-import
import pytest
from urlobject import URLObject
import responses
from nylas.client import APIClient
from nylas.client.restful_models import Contact
from nylas.utils import AuthMethod


def urls_equal(url1, url2):
    """
    Compare two URLObjects, without regard to the order of their query strings.
    """
    return (
        url1.without_query() == url2.without_query()
        and url1.query_dict == url2.query_dict
    )


def test_custom_client():
    # Can specify API server
    custom = APIClient(api_server="https://example.com")
    assert custom.api_server == "https://example.com"
    # Must be a valid URL
    with pytest.raises(Exception) as exc:
        APIClient(api_server="invalid")
    assert exc.value.args[0] == (
        "When overriding the Nylas API server address, " "you must include https://"
    )


@pytest.mark.usefixtures("mock_resources")
def test_client_access_token(api_client, mocked_responses):
    api_client.access_token = "foo"
    assert api_client.access_token == "foo"
    api_client.room_resources.first()
    assert mocked_responses.calls[0].request.headers["Authorization"] == "Bearer foo"
    api_client.access_token = "bar"
    api_client.room_resources.first()
    assert api_client.access_token == "bar"
    assert mocked_responses.calls[1].request.headers["Authorization"] == "Bearer bar"
    api_client.access_token = None
    api_client.room_resources.first()
    assert api_client.access_token is None
    assert "Authorization" not in api_client.session.headers


def test_client_headers():
    client = APIClient(client_id="whee", client_secret="foo")
    headers = client.session.headers
    assert headers["X-Nylas-API-Wrapper"] == "python"
    assert headers["X-Nylas-Client-Id"] == "whee"
    assert "Nylas-API-Version" in headers
    assert "Nylas Python SDK" in headers["User-Agent"]
    assert "Authorization" not in headers


def test_client_admin_headers():
    client = APIClient(client_id="bounce", client_secret="foo")
    headers = client.admin_session.headers
    assert headers["Authorization"] == "Basic Zm9vOg=="
    assert headers["X-Nylas-API-Wrapper"] == "python"
    assert headers["X-Nylas-Client-Id"] == "bounce"
    assert "Nylas-API-Version" in headers
    assert "Nylas Python SDK" in headers["User-Agent"]


def test_custom_api_version():
    # Can specify API server
    custom = APIClient(api_version="500")
    assert custom.api_version == "500"


def test_client_authentication_url(api_client, api_url):
    expected = (
        URLObject(api_url)
        .with_path("/oauth/authorize")
        .set_query_params(
            [
                ("login_hint", ""),
                ("state", ""),
                ("redirect_uri", "/redirect"),
                ("response_type", "code"),
                ("client_id", "None"),
                ("scopes", "email,calendar,contacts"),
            ]
        )
    )
    actual = URLObject(api_client.authentication_url("/redirect"))
    assert urls_equal(expected, actual)

    actual2 = URLObject(api_client.authentication_url("/redirect", login_hint="hint"))
    expected2 = expected.set_query_param("login_hint", "hint")
    assert urls_equal(expected2, actual2)

    actual3 = URLObject(api_client.authentication_url("/redirect", state="confusion"))
    expected3 = expected.set_query_param("state", "confusion")
    assert urls_equal(expected3, actual3)


def test_client_authentication_url_custom_scopes(api_client, api_url):
    expected = (
        URLObject(api_url)
        .with_path("/oauth/authorize")
        .set_query_params(
            [
                ("login_hint", ""),
                ("state", ""),
                ("redirect_uri", "/redirect"),
                ("response_type", "code"),
                ("client_id", "None"),
                ("scopes", "email"),
            ]
        )
    )
    actual = URLObject(api_client.authentication_url("/redirect", scopes="email"))
    assert urls_equal(expected, actual)

    actual2 = URLObject(
        api_client.authentication_url("/redirect", scopes=["calendar", "contacts"])
    )
    expected2 = expected.set_query_param("scopes", "calendar,contacts")
    assert urls_equal(expected2, actual2)


def test_client_authentication_url_scopes_none(api_client, api_url):
    expected = (
        URLObject(api_url)
        .with_path("/oauth/authorize")
        .set_query_params(
            [
                ("login_hint", ""),
                ("state", ""),
                ("redirect_uri", "/redirect"),
                ("response_type", "code"),
                ("client_id", "None"),
                # no scopes parameter
            ]
        )
    )
    actual = URLObject(api_client.authentication_url("/redirect", scopes=None))
    assert urls_equal(expected, actual)


def test_client_authentication_url_optional_params(api_client, api_url):
    expected = (
        URLObject(api_url)
        .with_path("/oauth/authorize")
        .set_query_params(
            [
                ("login_hint", ""),
                ("state", ""),
                ("redirect_uri", "/redirect"),
                ("response_type", "code"),
                ("client_id", "None"),
                ("scopes", "email"),
                ("provider", "gmail"),
                ("redirect_on_error", "false"),
            ]
        )
    )
    actual = URLObject(
        api_client.authentication_url(
            "/redirect", scopes="email", provider="gmail", redirect_on_error=False
        )
    )
    assert urls_equal(expected, actual)


def test_client_authentication_url_invalid_param_values(api_client, api_url):
    expected = (
        URLObject(api_url)
        .with_path("/oauth/authorize")
        .set_query_params(
            [
                ("login_hint", ""),
                ("state", ""),
                ("redirect_uri", "/redirect"),
                ("response_type", "code"),
                ("client_id", "None"),
                ("scopes", "email"),
            ]
        )
    )
    actual = URLObject(
        api_client.authentication_url("/redirect", scopes="email", provider="Google")
    )
    assert urls_equal(expected, actual)

    expected2 = expected.set_query_param("provider", "gmail")

    actual2 = URLObject(
        api_client.authentication_url(
            "/redirect", scopes="email", provider="gmail", redirect_on_error="true"
        )
    )

    assert urls_equal(expected2, actual2)


def test_client_token_for_code(mocked_responses, api_client, api_url):
    endpoint = re.compile(api_url + "/oauth/token")
    response_body = json.dumps({"access_token": "hooray"})
    mocked_responses.add(
        responses.POST,
        endpoint,
        content_type="application/json",
        status=200,
        body=response_body,
    )

    assert api_client.token_for_code("foo") == "hooray"
    assert len(mocked_responses.calls) == 1
    request = mocked_responses.calls[0].request
    body = parse_qs(request.body)
    assert body["grant_type"] == ["authorization_code"]
    assert body["code"] == ["foo"]


def test_client_opensource_api(api_client):
    # pylint: disable=singleton-comparison
    assert api_client.is_opensource_api() == True
    api_client.client_id = "foo"
    api_client.client_secret = "super-sekrit"
    assert api_client.is_opensource_api() == False
    api_client.client_id = api_client.client_secret = None
    assert api_client.is_opensource_api() == True


def test_client_revoke_token(mocked_responses, api_client, api_url):
    endpoint = re.compile(api_url + "/oauth/revoke")
    mocked_responses.add(responses.POST, endpoint, status=200, body="")

    api_client.auth_token = "foo"
    api_client.access_token = "bar"
    api_client.revoke_token()
    assert api_client.auth_token is None
    assert api_client.access_token is None
    assert len(mocked_responses.calls) == 1


def test_create_resources(mocked_responses, api_client, api_url):
    contacts_data = [
        {"id": 1, "name": "first", "email": "first@example.com"},
        {"id": 2, "name": "second", "email": "second@example.com"},
    ]
    mocked_responses.add(
        responses.POST,
        api_url + "/contacts",
        content_type="application/json",
        status=200,
        body=json.dumps(contacts_data),
    )

    post_data = list(contacts_data)  # make a copy
    for contact in post_data:
        del contact["id"]

    contacts = api_client._create_resources(Contact, post_data)
    assert len(contacts) == 2
    assert all(isinstance(contact, Contact) for contact in contacts)
    assert len(mocked_responses.calls) == 1


def test_call_resource_method(mocked_responses, api_client, api_url):
    contact_data = {"id": 1, "name": "first", "email": "first@example.com"}
    mocked_responses.add(
        responses.POST,
        api_url + "/contacts/1/remove_duplicates",
        content_type="application/json",
        status=200,
        body=json.dumps(contact_data),
    )

    contact = api_client._call_resource_method(Contact, 1, "remove_duplicates", {})
    assert isinstance(contact, Contact)
    assert len(mocked_responses.calls) == 1


def test_201_response(mocked_responses, api_client, api_url):
    contact_data = {"id": 1, "given_name": "Charlie", "surname": "Bucket"}
    mocked_responses.add(
        responses.POST,
        api_url + "/contacts",
        content_type="application/json",
        status=201,  # This HTTP status still indicates success,
        # even though it's not 200.
        body=json.dumps(contact_data),
    )
    contact = api_client.contacts.create()
    contact.save()
    assert len(mocked_responses.calls) == 1


def test_301_response(mocked_responses, api_client, api_url):
    contact_data = {"id": 1, "given_name": "Charlie", "surname": "Bucket"}
    mocked_responses.add(
        responses.GET,
        api_url + "/contacts/first",
        status=301,
        headers={"Location": api_url + "/contacts/1"},
    )
    mocked_responses.add(
        responses.GET,
        api_url + "/contacts/1",
        content_type="application/json",
        status=200,
        body=json.dumps(contact_data),
    )
    contact = api_client.contacts.get("first")
    assert contact["id"] == 1
    assert contact["given_name"] == "Charlie"
    assert contact["surname"] == "Bucket"
    assert len(mocked_responses.calls) == 2


def test_pagination(mocked_responses, api_client, api_url):
    def callback(request):
        url = URLObject(request.url)
        limit = int(url.query_dict.get("limit") or 50)
        offset = int(url.query_dict.get("offset") or 0)
        fake_data = [{"id": i} for i in range(offset, limit + offset)]
        return (200, {}, json.dumps(fake_data))

    mocked_responses.add_callback(
        responses.GET,
        "{base}/contacts".format(base=api_url),
        content_type="application/json",
        callback=callback,
    )

    contacts = list(api_client.contacts.where(limit=75))
    assert len(contacts) == 75


def test_count(mocked_responses, api_client, api_url):
    count_data = {"count": 721}
    mocked_responses.add(
        responses.GET,
        api_url + "/contacts",
        content_type="application/json",
        body=json.dumps(count_data),
    )

    contact_count = api_client.contacts.count()
    assert contact_count == 721


def test_add_auth_header_bearer(api_client):
    api_client.access_token = "access_token"
    auth_header = api_client._add_auth_header(AuthMethod.BEARER)
    assert auth_header == {"Authorization": "Bearer access_token"}


def test_add_auth_header_basic_client_id_and_secret(api_client):
    api_client.client_id = "client_id"
    api_client.client_secret = "client_secret"
    auth_header = api_client._add_auth_header(AuthMethod.BASIC_CLIENT_ID_AND_SECRET)
    assert auth_header == {"Authorization": "Basic Y2xpZW50X2lkOmNsaWVudF9zZWNyZXQ="}


def test_add_auth_header_basic(api_client):
    api_client.client_secret = "client_secret"
    auth_header = api_client._add_auth_header(AuthMethod.BASIC)
    assert auth_header == {"Authorization": "Basic Y2xpZW50X3NlY3JldDo="}



from datetime import datetime

import pytest
from urlobject import URLObject
from nylas.client.restful_models import Message, Draft, Label
from nylas.utils import timestamp_from_dt


@pytest.mark.usefixtures("mock_threads")
def test_thread_attrs(api_client):
    thread = api_client.threads.first()
    expected_first = datetime(2016, 1, 2, 3, 4, 5)
    expected_last = datetime(2017, 1, 2, 3, 4, 5)
    expected_last_received = datetime(2017, 1, 2, 3, 4, 5)
    expected_last_sent = datetime(2017, 1, 1, 1, 1, 1)

    assert thread.first_message_timestamp == timestamp_from_dt(expected_first)
    assert thread.first_message_at == expected_first
    assert thread.last_message_timestamp == timestamp_from_dt(expected_last)
    assert thread.last_message_at == expected_last
    assert thread.last_message_received_timestamp == timestamp_from_dt(
        expected_last_received
    )
    assert thread.last_message_received_at == expected_last_received
    assert thread.last_message_sent_timestamp == timestamp_from_dt(expected_last_sent)
    assert thread.last_message_sent_at == expected_last_sent


def test_update_thread_attrs(api_client):
    thread = api_client.threads.create()
    first = datetime(2017, 2, 3, 10, 0, 0)
    second = datetime(2016, 10, 5, 14, 30, 0)
    # timestamps and datetimes are handled totally separately
    thread.last_message_at = first
    thread.last_message_timestamp = timestamp_from_dt(second)
    assert thread.last_message_at == first
    assert thread.last_message_timestamp == timestamp_from_dt(second)
    # but datetimes overwrite timestamps when serializing to JSON
    assert thread.as_json()["last_message_timestamp"] == timestamp_from_dt(first)


@pytest.mark.usefixtures("mock_threads")
def test_thread_folder(api_client):
    thread = api_client.threads.first()
    assert len(thread.labels) == 0  # pylint: disable=len-as-condition
    assert len(thread.folders) == 1
    assert thread.folders[0].display_name == "Inbox"
    assert not thread.unread
    assert thread.starred


@pytest.mark.usefixtures("mock_folder_account", "mock_threads", "mock_thread")
def test_thread_change(api_client):
    thread = api_client.threads.first()

    assert thread.starred
    thread.unstar()
    assert not thread.starred
    thread.star()
    assert thread.starred

    thread.update_folder("qwer")
    assert len(thread.folders) == 1
    assert thread.folders[0].id == "qwer"


@pytest.mark.usefixtures("mock_threads", "mock_messages")
def test_thread_messages(api_client):
    thread = api_client.threads.first()
    assert thread.messages
    assert all(isinstance(message, Message) for message in thread.messages)


@pytest.mark.usefixtures("mock_threads", "mock_drafts")
def test_thread_drafts(api_client):
    thread = api_client.threads.first()
    assert thread.drafts
    assert all(isinstance(draft, Draft) for draft in thread.drafts)


@pytest.mark.usefixtures("mock_labelled_thread", "mock_labels")
def test_thread_label(api_client):
    thread = api_client.threads.get(111)
    assert len(thread.labels) == 2
    assert all(isinstance(label, Label) for label in thread.labels)

    returned = thread.add_label("fake1")
    assert len(thread.labels) == 3
    assert thread.labels == returned

    returned = thread.remove_label("fake1")
    assert len(thread.labels) == 2  # pylint: disable=len-as-condition
    assert thread.labels == returned


@pytest.mark.usefixtures("mock_labelled_thread", "mock_labels")
def test_thread_labels(api_client):
    thread = api_client.threads.get(111)
    assert len(thread.labels) == 2
    assert all(isinstance(label, Label) for label in thread.labels)

    returned = thread.add_labels(["fake1", "fake2"])
    assert len(thread.labels) == 4
    assert thread.labels == returned

    label_ids = [l.id for l in thread.labels]
    returned = thread.remove_labels(label_ids)
    assert len(thread.labels) == 0  # pylint: disable=len-as-condition
    assert thread.labels == returned


@pytest.mark.usefixtures("mock_threads", "mock_thread")
def test_thread_read(api_client):
    thread = api_client.threads.first()
    assert thread.unread is False
    thread.mark_as_unread()
    assert thread.unread is True
    thread.mark_as_read()
    assert thread.unread is False
    # mark_as_seen() is a synonym for mark_as_read()
    thread.mark_as_unread()
    assert thread.unread is True
    thread.mark_as_seen()
    assert thread.unread is False


@pytest.mark.usefixtures("mock_threads")
def test_thread_reply(api_client):
    thread = api_client.threads.first()
    draft = thread.create_reply()
    assert isinstance(draft, Draft)
    assert draft.thread_id == thread.id
    assert draft.subject == thread.subject


@pytest.mark.usefixtures("mock_threads")
def test_filter_threads_dt(mocked_responses, api_client):
    api_client.threads.where(started_before=datetime(2010, 6, 1)).all()
    assert len(mocked_responses.calls) == 1
    request = mocked_responses.calls[0].request
    url = URLObject(request.url)
    assert url.query_dict["started_before"] == "1275350400"


@pytest.mark.usefixtures("mock_threads")
def test_filter_threads_ts(mocked_responses, api_client):
    api_client.threads.where(started_before=1275350400).all()
    assert len(mocked_responses.calls) == 1
    request = mocked_responses.calls[0].request
    url = URLObject(request.url)
    assert url.query_dict["started_before"] == "1275350400"



import json

from requests import ReadTimeout

from nylas.client.delta_models import Delta, Deltas


class DeltaCollection:
    path = "delta"

    def __init__(self, api):
        self.api = api

    def latest_cursor(self):
        """
        Returns the latest delta cursor

        Returns:
            str: The latest cursor

        Raises:
            RuntimeError: If the server returns an object without a cursor
        """

        response = self.api._post_resource(
            Delta, "latest_cursor", None, None, path=self.path
        )
        if "cursor" not in response:
            raise RuntimeError(
                "Unexpected response from the API server. Returned 200 but no 'cursor' string found."
            )

        return response["cursor"]

    def since(self, cursor, view=None, include_types=None, excluded_types=None):
        """
        Get a list of delta cursors since a specified cursor

        Args:
            cursor (str): The first cursor to request from
            view (str): Value representing if delta expands thread and message objects.
            include_types (list[str] | str): The objects to exclusively include in the returned deltas. Note you cannot set both included and excluded types.
            excluded_types (list[str] | str): The objects to exclude in the returned deltas. Note you cannot set both included and excluded types.

        Returns:
            Deltas: The API response containing the list of deltas

        Raises:
            ValueError: If both include_types and excluded_types are set
        """

        include_types, excluded_types = _validate_types(include_types, excluded_types)
        response = self.api._get_resource_raw(
            Delta,
            None,
            path=self.path,
            cursor=cursor,
            view=view,
            include_types=include_types,
            excluded_types=excluded_types,
        ).json()
        return Deltas.create(self.api, **response)

    def stream(
        self,
        cursor,
        callback=None,
        timeout=None,
        view=None,
        include_types=None,
        excluded_types=None,
    ):
        """
        Stream deltas

        Args:
            cursor (str): The cursor to stream from
            callback: A callable function to invoke on each delta received. No callback is set by default.
            timeout (int): The number of seconds to stream for before timing out. No timeout is set by default.
            view (str): Value representing if delta expands thread and message objects.
            include_types (list[str] | str): The objects to exclusively include in the returned deltas. Note you cannot set both included and excluded types.
            excluded_types (list[str] | str): The objects to exclude in the returned deltas. Note you cannot set both included and excluded types.

        Returns:
            list[Delta]: The list of streamed deltas

        Raises:
            ValueError: If both include_types and excluded_types are set
        """

        deltas = []
        include_types, excluded_types = _validate_types(include_types, excluded_types)
        emit_deltas = False
        if callback and callable(callback):
            emit_deltas = True

        try:
            response = self.api._get_resource_raw(
                Delta,
                "streaming",
                stream=True,
                path=self.path,
                stream_timeout=timeout,
                cursor=cursor,
                view=view,
                include_types=include_types,
                excluded_types=excluded_types,
            )
            for raw_rsp in response.iter_lines():
                if raw_rsp:
                    response_json = json.loads(raw_rsp)
                    delta = Delta.create(self.api, **response_json)
                    deltas.append(delta)
                    if emit_deltas:
                        callback(delta)
        except ReadTimeout:
            pass

        return deltas

    def longpoll(
        self,
        cursor,
        timeout,
        callback=None,
        view=None,
        include_types=None,
        excluded_types=None,
    ):
        """
        Long-poll for deltas

        Args:
            cursor (str): The cursor to poll from
            timeout (int): The number of seconds to poll for before timing out
            callback: A callable function to invoke on each delta received. No callback is set by default.
            view (str): Value representing if delta expands thread and message objects.
            include_types (list[str] | str): The objects to exclusively include in the returned deltas. Note you cannot set both included and excluded types.
            excluded_types (list[str] | str): The objects to exclude in the returned deltas. Note you cannot set both included and excluded types.

        Returns:
            Deltas: The API response containing the list of deltas

        Raises:
            ValueError: If both include_types and excluded_types are set
        """

        delta = {}
        include_types, excluded_types = _validate_types(include_types, excluded_types)
        emit_deltas = False
        if callback and callable(callback):
            emit_deltas = True

        buffer = bytearray()
        response = self.api._get_resource_raw(
            Delta,
            "longpoll",
            stream=True,
            path=self.path,
            timeout=timeout,
            cursor=cursor,
            view=view,
            include_types=include_types,
            excluded_types=excluded_types,
        )
        for raw_rsp in response.iter_lines():
            if raw_rsp:
                buffer.extend(raw_rsp)
                try:
                    buffer_json = json.loads(buffer.decode())
                    delta = Deltas.create(self.api, **buffer_json)
                    if emit_deltas:
                        callback(delta)
                except ValueError:
                    continue

        return delta


# Helper functions for validating type inputs
def _validate_types(include_types, excluded_types):
    if include_types and excluded_types:
        raise ValueError("You cannot set both include_types and excluded_types")

    return _join_types(include_types), _join_types(excluded_types)


def _join_types(types):
    if types:
        if isinstance(types, str):
            return types
        try:
            return ",".join(types)
        except TypeError:
            return None



from datetime import datetime
import json

import six
import pytest
from urlobject import URLObject
from nylas.client.restful_models import Message
from nylas.utils import timestamp_from_dt
import pytz


@pytest.mark.usefixtures("mock_messages")
def test_messages(api_client):
    message = api_client.messages.first()
    assert len(message.labels) == 1
    assert message.labels[0].display_name == "Inbox"
    assert message.folder is None
    assert message.unread
    assert not message.starred


@pytest.mark.usefixtures("mock_messages")
def test_message_attrs(api_client):
    message = api_client.messages.first()
    expected_received = datetime(2010, 2, 2, 2, 22, 22)
    assert message.received_at == expected_received
    assert message.date == timestamp_from_dt(expected_received)


@pytest.mark.usefixtures("mock_account", "mock_messages", "mock_message")
def test_message_stars(api_client):
    message = api_client.messages.first()
    assert message.starred is False
    message.star()
    assert message.starred is True
    message.unstar()
    assert message.starred is False


@pytest.mark.usefixtures("mock_account", "mock_messages", "mock_message")
def test_message_read(api_client):
    message = api_client.messages.first()
    assert message.unread is True
    message.mark_as_read()
    assert message.unread is False
    message.mark_as_unread()
    assert message.unread is True
    # mark_as_seen() is a synonym for mark_as_read()
    message.mark_as_seen()
    assert message.unread is False


@pytest.mark.usefixtures("mock_account", "mock_messages", "mock_message")
def test_message_labels(api_client):
    message = api_client.messages.first()
    message.add_label("fghj")
    msg_labels = [l.id for l in message.labels]
    assert "abcd" in msg_labels
    assert "fghj" in msg_labels
    message.remove_label("fghj")
    msg_labels = [l.id for l in message.labels]
    assert "abcd" in msg_labels
    assert "fghj" not in msg_labels

    # Test that folders don't do anything when labels are in effect
    message.update_folder("zxcv")
    assert message.folder is None


@pytest.mark.usefixtures("mock_account", "mock_message", "mock_messages")
def test_message_raw(api_client, account_id):
    message = api_client.messages.first()
    raw = message.raw
    assert isinstance(raw, six.binary_type)
    parsed = json.loads(raw)
    assert parsed == {
        "object": "message",
        "to": [{"email": "foo@yahoo.com", "name": "Foo"}],
        "from": [{"email": "bar@gmail.com", "name": "Bar"}],
        "account_id": account_id,
        "labels": [{"display_name": "Inbox", "name": "inbox", "id": "abcd"}],
        "starred": False,
        "unread": True,
        "id": "1234",
        "subject": "Test Message",
    }


@pytest.mark.usefixtures("mock_message")
def test_message_delete_by_id(mocked_responses, api_client):
    api_client.messages.delete(1234, forceful=True)
    assert len(mocked_responses.calls) == 1
    request = mocked_responses.calls[0].request
    url = URLObject(request.url)
    assert url.query_dict["forceful"] == "True"


@pytest.mark.usefixtures("mock_message")
def test_message_resolution(mocked_responses, api_client, account_id):
    message = api_client.messages.get(1234)
    assert message.object == "message"
    assert message.to == [{"email": "foo@yahoo.com", "name": "Foo"}]
    assert message.from_ == [{"email": "bar@gmail.com", "name": "Bar"}]
    assert message["from"] == [{"email": "bar@gmail.com", "name": "Bar"}]
    assert message.account_id == account_id
    assert message._labels == [{"display_name": "Inbox", "name": "inbox", "id": "abcd"}]
    assert message.id == "1234"
    assert message.subject == "Test Message"
    assert message.starred is False
    assert message.unread is True


@pytest.mark.usefixtures("mock_messages")
def test_slice_messages(api_client):
    messages = api_client.messages[0:2]
    assert len(messages) == 3
    assert all(isinstance(message, Message) for message in messages)


@pytest.mark.usefixtures("mock_messages")
def test_filter_messages_dt(mocked_responses, api_client):
    api_client.messages.where(received_before=datetime(2010, 6, 1)).all()
    assert len(mocked_responses.calls) == 1
    request = mocked_responses.calls[0].request
    url = URLObject(request.url)
    assert url.query_dict["received_before"] == "1275350400"


@pytest.mark.usefixtures("mock_messages")
def test_filter_messages_dt_with_timezone(mocked_responses, api_client):
    api_client.messages.where(
        received_before=datetime(2010, 6, 1, tzinfo=pytz.utc)
    ).all()
    assert len(mocked_responses.calls) == 1
    request = mocked_responses.calls[0].request
    url = URLObject(request.url)
    assert url.query_dict["received_before"] == "1275350400"


@pytest.mark.usefixtures("mock_messages")
def test_filter_messages_ts(mocked_responses, api_client):
    api_client.messages.where(received_before=1275350400).all()
    assert len(mocked_responses.calls) == 1
    request = mocked_responses.calls[0].request
    url = URLObject(request.url)
    assert url.query_dict["received_before"] == "1275350400"


@pytest.mark.usefixtures("mock_message", "mock_messages")
def test_message_metadata(mocked_responses, api_client):
    message = api_client.messages.first()
    message["metadata"] = {"test": "value"}
    message.save()
    assert message.metadata == {"test": "value"}



import json

import pytest
from urlobject import URLObject

from nylas.client.authentication_models import Authentication, Integration, Grant


@pytest.mark.usefixtures("mock_integrations")
def test_authentication_api_url(mocked_responses, api_client):
    authentication = api_client.authentication
    integrations = authentication.integrations
    integrations.first()
    request = mocked_responses.calls[0].request
    assert URLObject(request.url).hostname == "beta.us.nylas.com"
    authentication.app_name = "test_app"
    integrations.first()
    request = mocked_responses.calls[1].request
    assert URLObject(request.url).hostname == "test_app.us.nylas.com"
    authentication.region = Authentication.Region.EU
    integrations.first()
    request = mocked_responses.calls[2].request
    assert URLObject(request.url).hostname == "test_app.eu.nylas.com"


@pytest.mark.usefixtures("mock_integrations")
def test_integration(mocked_responses, api_client):
    integration = api_client.authentication.integrations.first()
    request = mocked_responses.calls[0].request
    assert URLObject(request.url).path == "/connect/integrations"
    assert request.method == "GET"
    assert isinstance(integration, Integration)
    assert integration.name == "Nylas Playground"
    assert integration.id == "zoom"
    assert integration.provider == "zoom"
    assert integration.settings["client_id"] == "test_client_id"
    assert integration.settings["client_secret"] == "test_client_secret"
    assert integration.redirect_uris[0] == "https://www.nylas.com"
    assert integration.expires_in == 12000


@pytest.mark.usefixtures("mock_integrations")
def test_single_integration(mocked_responses, api_client):
    integration = api_client.authentication.integrations.get(
        Authentication.Provider.ZOOM
    )
    request = mocked_responses.calls[0].request
    assert URLObject(request.url).path == "/connect/integrations/zoom"
    assert request.method == "GET"
    assert isinstance(integration, Integration)
    assert integration.id == "zoom"
    assert integration.provider == "zoom"


@pytest.mark.usefixtures("mock_integrations")
def test_update_integration(mocked_responses, api_client):
    integration = api_client.authentication.integrations.get(
        Authentication.Provider.ZOOM
    )
    integration.name = "Updated Integration Name"
    integration.save()
    assert len(mocked_responses.calls) == 2
    request = mocked_responses.calls[1].request
    assert URLObject(request.url).path == "/connect/integrations/zoom"
    assert request.method == "PATCH"
    assert json.loads(request.body) == {
        "name": "Updated Integration Name",
        "settings": {
            "client_id": "test_client_id",
            "client_secret": "test_client_secret",
        },
        "redirect_uris": ["https://www.nylas.com"],
        "expires_in": 12000,
        "scope": [],
    }
    assert isinstance(integration, Integration)
    assert integration.id == "zoom"
    assert integration.provider == "zoom"
    assert integration.name == "Updated Integration Name"


@pytest.mark.usefixtures("mock_integrations")
def test_delete_integration(mocked_responses, api_client):
    api_client.authentication.integrations.delete(Authentication.Provider.ZOOM)
    request = mocked_responses.calls[0].request
    assert URLObject(request.url).path == "/connect/integrations/zoom"
    assert request.method == "DELETE"


@pytest.mark.usefixtures("mock_integrations")
def test_create_integration(mocked_responses, api_client):
    integration = api_client.authentication.integrations.create()
    integration.name = "Nylas Playground"
    integration.provider = Authentication.Provider.ZOOM
    integration.settings["client_id"] = "test_client_id"
    integration.settings["client_secret"] = "test_client_secret"
    integration.redirect_uris = ["https://www.nylas.com"]
    integration.expires_in = 12000
    integration.scope = ["test.scope"]
    integration.save()
    request = mocked_responses.calls[0].request
    assert URLObject(request.url).path == "/connect/integrations"
    assert request.method == "POST"
    assert json.loads(request.body) == {
        "name": "Nylas Playground",
        "provider": "zoom",
        "settings": {
            "client_id": "test_client_id",
            "client_secret": "test_client_secret",
        },
        "redirect_uris": ["https://www.nylas.com"],
        "expires_in": 12000,
        "scope": ["test.scope"],
    }
    assert isinstance(integration, Integration)
    assert integration.name == "Nylas Playground"
    assert integration.id == "zoom"
    assert integration.provider == "zoom"
    assert integration.settings["client_id"] == "test_client_id"
    assert integration.settings["client_secret"] == "test_client_secret"
    assert integration.redirect_uris[0] == "https://www.nylas.com"
    assert integration.expires_in == 12000
    assert integration.scope == ["test.scope"]


@pytest.mark.usefixtures("mock_grants")
def test_grant(mocked_responses, api_client):
    grant = api_client.authentication.grants.first()
    request = mocked_responses.calls[0].request
    assert URLObject(request.url).path == "/connect/grants"
    assert request.method == "GET"
    assert isinstance(grant, Grant)
    assert grant.id == "grant-id"
    assert grant.provider == "zoom"
    assert grant.grant_status == "valid"
    assert grant.email == "email@example.com"
    assert grant.metadata == {"isAdmin": True}
    assert grant.scope[0] == "meeting:write"
    assert grant.user_agent == "string"
    assert grant.ip == "string"
    assert grant.created_at == 1617817109
    assert grant.updated_at == 1617817109


@pytest.mark.usefixtures("mock_grants")
def test_single_grant(mocked_responses, api_client):
    grant = api_client.authentication.grants.get("grant-id")
    request = mocked_responses.calls[0].request
    assert URLObject(request.url).path == "/connect/grants/grant-id"
    assert request.method == "GET"
    assert isinstance(grant, Grant)
    assert grant.id == "grant-id"


@pytest.mark.usefixtures("mock_grants")
def test_update_grant(mocked_responses, api_client):
    grant = api_client.authentication.grants.get("grant-id")
    grant.settings = {"refresh_token": "test_token"}
    grant.save()
    assert len(mocked_responses.calls) == 2
    request = mocked_responses.calls[1].request
    assert URLObject(request.url).path == "/connect/grants/grant-id"
    assert request.method == "PATCH"
    assert json.loads(request.body) == {
        "settings": {"refresh_token": "test_token"},
        "metadata": {"isAdmin": True},
        "scope": ["meeting:write"],
    }
    assert isinstance(grant, Grant)
    assert grant.id == "grant-id"
    assert grant.settings == {"refresh_token": "test_token"}


@pytest.mark.usefixtures("mock_grants")
def test_delete_grant(mocked_responses, api_client):
    api_client.authentication.grants.delete("grant-id")
    request = mocked_responses.calls[0].request
    assert URLObject(request.url).path == "/connect/grants/grant-id"
    assert request.method == "DELETE"


@pytest.mark.usefixtures("mock_grants")
def test_create_grant(mocked_responses, api_client):
    grant = api_client.authentication.grants.create()
    grant.provider = Authentication.Provider.ZOOM
    grant.settings = {"refresh_token": "test-refresh-token"}
    grant.metadata = {"isAdmin": True}
    grant.scope = ["meeting:write"]
    grant.save()
    request = mocked_responses.calls[0].request
    assert URLObject(request.url).path == "/connect/grants"
    assert request.method == "POST"
    assert json.loads(request.body) == {
        "provider": "zoom",
        "settings": {"refresh_token": "test-refresh-token"},
        "scope": ["meeting:write"],
        "metadata": {"isAdmin": True},
    }
    assert isinstance(grant, Grant)


@pytest.mark.usefixtures("mock_grants")
def test_grant_on_demand_sync(mocked_responses, api_client):
    grant = api_client.authentication.grants.on_demand_sync("grant-id", sync_from=12000)
    request = mocked_responses.calls[0].request
    assert request.path_url == "/connect/grants/grant-id/sync?sync_from=12000"
    assert request.method == "POST"
    assert request.body is None
    assert isinstance(grant, Grant)


@pytest.mark.usefixtures("mock_authentication_hosted_auth")
def test_grant_authentication_hosted_auth(mocked_responses, api_client):
    api_client.authentication.hosted_authentication(
        provider=Authentication.Provider.ZOOM,
        redirect_uri="https://myapp.com/callback-handler",
        grant_id="test-grant-id",
        login_hint="example@email.com",
        state="test-state",
        expires_in=60,
        settings={"refresh_token": "test-refresh-token"},
        metadata={"isAdmin": True},
        scope=["meeting:write"],
    )
    request = mocked_responses.calls[0].request
    assert URLObject(request.url).path == "/connect/auth"
    assert request.method == "POST"
    assert json.loads(request.body) == {
        "provider": "zoom",
        "redirect_uri": "https://myapp.com/callback-handler",
        "grant_id": "test-grant-id",
        "login_hint": "example@email.com",
        "state": "test-state",
        "expires_in": 60,
        "settings": {"refresh_token": "test-refresh-token"},
        "scope": ["meeting:write"],
        "metadata": {"isAdmin": True},
    }


@pytest.mark.usefixtures("mock_authentication_hosted_auth")
def test_grant_authentication_hosted_auth_enhanced_events(mocked_responses, api_client):
    api_client.authentication._hosted_authentication_enhanced_events(
        provider=Authentication.Provider.ZOOM,
        redirect_uri="https://myapp.com/callback-handler",
        account_id="test-account-id",
    )
    request = mocked_responses.calls[0].request
    assert URLObject(request.url).path == "/connect/auth"
    assert request.method == "POST"
    assert json.loads(request.body) == {
        "provider": "zoom",
        "redirect_uri": "https://myapp.com/callback-handler",
        "account_id": "test-account-id",
    }



from datetime import datetime

import pytest

from nylas.client.restful_models import Component


def blank_component(api_client):
    component = api_client.components.create()
    component.name = "Python Component Test"
    component.type = "agenda"
    component.public_account_id = "test-account-id"
    component.access_token = "test-access-token"
    return component


@pytest.mark.usefixtures("mock_components")
def test_components(api_client):
    component = api_client.components.first()
    assert isinstance(component, Component)
    assert component.id == "component-id"
    assert component.active is True
    assert component.name == "PyTest Component"
    assert component.public_account_id == "account-id"
    assert component.public_application_id == "application-id"
    assert component.type == "agenda"
    assert component.created_at == datetime.strptime(
        "2021-10-22T18:02:10.000Z", "%Y-%m-%dT%H:%M:%S.%fZ"
    )
    assert component.updated_at == datetime.strptime(
        "2021-10-22T18:02:10.000Z", "%Y-%m-%dT%H:%M:%S.%fZ"
    )
    assert component.public_token_id == "token-id"


@pytest.mark.usefixtures("mock_components")
def test_components(api_client):
    component = api_client.components.first()
    assert isinstance(component, Component)
    assert component.id == "component-id"
    assert component.active is True
    assert component.name == "PyTest Component"
    assert component.public_account_id == "account-id"
    assert component.public_application_id == "application-id"
    assert component.type == "agenda"
    assert component.created_at == datetime.strptime(
        "2021-10-22T18:02:10.000Z", "%Y-%m-%dT%H:%M:%S.%fZ"
    )
    assert component.updated_at == datetime.strptime(
        "2021-10-22T18:02:10.000Z", "%Y-%m-%dT%H:%M:%S.%fZ"
    )
    assert component.public_token_id == "token-id"


@pytest.mark.usefixtures("mock_components_create_response")
def test_create_components(api_client):
    component = blank_component(api_client)
    component.save()
    assert component.id == "cv4ei7syx10uvsxbs21ccsezf"


@pytest.mark.usefixtures("mock_components_create_response")
def test_modify_components(api_client):
    component = blank_component(api_client)
    component.id = "cv4ei7syx10uvsxbs21ccsezf"
    component.name = "Updated Name"
    component.save()
    assert component.name == "Updated Name"


@pytest.mark.usefixtures("mock_components_create_response")
def test_components_as_json_read_only(api_client):
    component = blank_component(api_client)
    component.id = "test-id"
    json = component.as_json()
    assert "id" not in json
    assert "public_application_id" not in json
    assert "created_at" not in json
    assert "updated_at" not in json



import sys
import re
import subprocess
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand


VERSION = ""
with open("nylas/_client_sdk_version.py", "r") as fd:
    VERSION = re.search(
        r'^__VERSION__\s*=\s*[\'"]([^\'"]*)[\'"]', fd.read(), re.MULTILINE
    ).group(1)

RUN_DEPENDENCIES = [
    "requests[security]>=2.4.2",
    "six>=1.4.1",
    "urlobject",
    "enum34>=1.1.10",
]
TEST_DEPENDENCIES = [
    "pytest",
    "pytest-cov",
    "pytest-timeout",
    "responses==0.10.5",
    "twine",
    "pytz",
]
RELEASE_DEPENDENCIES = ["bumpversion>=0.5.0", "twine>=3.4.2"]


class PyTest(TestCommand):
    user_options = [("pytest-args=", "a", "Arguments to pass to pytest")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        # pylint: disable=attribute-defined-outside-init
        self.pytest_args = [
            "--cov",
            "--cov-report=xml",
            "--junitxml",
            "./tests/output",
            "tests/",
        ]
        self.lint = False

    def finalize_options(self):
        TestCommand.finalize_options(self)
        # pylint: disable=attribute-defined-outside-init
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest

        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


def main():
    # A few handy release helpers.
    # For publishing you should install the extra 'release' dependencies
    # by running: pip install nylas['release']
    if len(sys.argv) > 1:
        if sys.argv[1] == "publish":
            try:
                subprocess.check_output(["git", "push", "--follow-tags"])
                subprocess.check_output(["python", "setup.py", "sdist"])
                subprocess.check_output(["twine", "upload", "-r", "testpypi", "dist/*"])
                subprocess.check_output(["twine", "upload", "dist/*"])
            except FileNotFoundError as e:
                print("Error encountered: {}.\n\n".format(e))
            sys.exit()
        elif sys.argv[1] == "release":
            if len(sys.argv) < 3:
                type_ = "patch"
            else:
                type_ = sys.argv[2]
            try:
                subprocess.check_output(
                    ["bumpversion", "--current-version", VERSION, type_]
                )
            except FileNotFoundError as e:
                print(
                    "Error encountered: {}.\n\n".format(e),
                    "Did you install the extra 'release' dependencies? (pip install nylas['release'])",
                )
            sys.exit()

    setup(
        name="nylas",
        version=VERSION,
        packages=find_packages(),
        install_requires=RUN_DEPENDENCIES,
        dependency_links=[],
        tests_require=TEST_DEPENDENCIES,
        extras_require={"test": TEST_DEPENDENCIES, "release": RELEASE_DEPENDENCIES},
        cmdclass={"test": PyTest},
        author="Nylas Team",
        author_email="support@nylas.com",
        description="Python bindings for Nylas, the next-generation email platform.",
        license="MIT",
        keywords="inbox app appserver email nylas contacts calendar",
        url="https://github.com/nylas/nylas-python",
        long_description_content_type="text/markdown",
        long_description="""
# Nylas REST API Python bindings
![Build Status](https://github.com/nylas/nylas-python/workflows/Test/badge.svg)
[![Code Coverage](https://codecov.io/gh/nylas/nylas-python/branch/main/graph/badge.svg)](https://codecov.io/gh/nylas/nylas-python)

Python bindings for the Nylas REST API. https://www.nylas.com/docs

The Nylas APIs power applications with email, calendar, and contacts CRUD and bi-directional sync from any inbox in the world.

Nylas is compatible with 100% of email service providers, so you only have to integrate once.
No more headaches building unique integrations against archaic and outdated IMAP and SMTP protocols.""",
    )


if __name__ == "__main__":
    main()



import pytest


@pytest.mark.usefixtures("mock_thread_search_response")
def test_search_threads(api_client):
    threads = api_client.threads.search("Helena")
    assert len(threads) == 1
    assert "Helena" in threads[0].snippet


@pytest.mark.usefixtures("mock_message_search_response")
def test_search_messages(api_client):
    messages = api_client.messages.search("Pinot")
    assert len(messages) == 2
    assert "Pinot" in messages[0].snippet
    assert "Pinot" in messages[1].snippet


@pytest.mark.usefixtures("mock_message_search_response")
def test_search_messages_with_limit_offset(mocked_responses, api_client):
    api_client.messages.search("Pinot", limit=10, offset=0)
    request = mocked_responses.calls[0].request
    assert request.path_url == "/messages/search?q=Pinot&limit=10&offset=0"


@pytest.mark.usefixtures("mock_message_search_response")
def test_search_drafts(api_client):
    with pytest.raises(Exception):
        api_client.drafts.search("Pinot")
